import logging
from inspect import stack
from pathlib import Path

import click
from click import BadParameter

from com.enovation.helper.click.python_literal_argument_and_option import PythonLiteralOption
from com.enovation.toolbox.excel_dashboard import ExcelDashboarder


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dashboard')
@click.pass_context
@click.argument('alias-config', type=str)
@click.argument('file', type=click.Path(exists=False))
@click.option(
    '--parameters', cls=PythonLiteralOption, type=dict, default=None,
    help='The parameters, provided as a dictionary, that are required in the dashboard configuration.'
)
@click.option(
    '--alias-parameters', cls=PythonLiteralOption, type=dict, default=None,
    help='The parameters, loaded as an alias in context, that are required in the dashboard configuration.'
)
def dashboard(ctx_context, alias_config, file, parameters, alias_parameters):
    """
    TODO

    The following business rules are checked:
    - BR_001, ALIAS_CONFIG should exist in context
    - BR_002, ALIAS_CONFIG should be a dictionary
    - BR_003, PARAMETERS, if provided, should be a dictionary
    - BR_004, ALIAS_PARAMETERS, if provided, should be a dictionary
    - BR_005, ALIAS_PARAMETERS, if provided, should only contain string values
    - BR_006, ALIAS_PARAMETERS, if provided, should list parameters that all exist in context
    - BR_007, there should not be duplicates across PARAMETERS and ALIAS_PARAMETERS

    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")
    _dict_the_config: dict
    _dict_the_parameters: dict = {}

    # BR_001, ALIAS_CONFIG should exist in context
    if alias_config not in ctx_context.obj:
        raise BadParameter(
            message=f"BR_001, no object labelled '{alias_config}' could be found in context, which is not expected.",
            param_hint=f"BR_001, ensure you effectively have an object labelled '{alias_config}' in context.",
            ctx=ctx_context
        )

    # BR_002, ALIAS_CONFIG should be a dictionary
    if not isinstance(ctx_context.obj[alias_config], dict):
        raise BadParameter(
            message=f"BR_002, object labelled '{alias_config}' is of type '{type(ctx_context.obj[alias_config])}', "
                    f"which is not expected.",
            param_hint=f"BR_002, ensure you provide a 'dict' instance.",
            ctx=ctx_context
        )

    _dict_the_config = ctx_context.obj[alias_config]

    if parameters is not None:

        # BR_003, PARAMETERS, if provided, should be a dictionary
        if not isinstance(parameters, dict):
            raise BadParameter(
                message=f"BR_003, parameter 'parameters' is of type '{type(parameters)}', "
                        f"which is not expected.",
                param_hint=f"BR_003, ensure you provide a parameter 'parameters' of type 'dict'.",
                ctx=ctx_context
            )

        _dict_the_parameters = parameters

    if alias_parameters is not None:

        # BR_004, ALIAS_PARAMETERS, if provided, should be a dictionary
        if not isinstance(alias_parameters, dict):
            raise BadParameter(
                message=f"BR_004, parameter 'parameters' is of type '{type(alias_parameters)}', "
                        f"which is not expected.",
                param_hint=f"BR_004, ensure you provide a parameter 'alias_parameters' of type 'dict'.",
                ctx=ctx_context
            )

        for k, v in alias_parameters.items():

            # BR_005, ALIAS_PARAMETERS, if provided, should only contain string values
            if not isinstance(v, str):
                raise BadParameter(
                    message=f"BR_005, alias_parameter with key '{k}' points to an alias label of type "
                            f"'{type(v)}', which is not expected.",
                    param_hint=f"BR_005, ensure you provide 'alias_parameters' of type 'dict[str,str]'.",
                    ctx=ctx_context
                )

            # BR_006, ALIAS_PARAMETERS, if provided, should list parameters that all exist in context
            if v not in ctx_context.obj:
                raise BadParameter(
                    message=f"BR_006, no object labelled '{v}' could be found in context, which is not expected.",
                    param_hint=f"BR_006, ensure you effectively have an object labelled '{v}' in context.",
                    ctx=ctx_context
                )

            # BR_007, there should not be duplicates across PARAMETERS and ALIAS_PARAMETERS
            if k in _dict_the_parameters:
                raise Exception(f"BR_007, parameter '{k}' is provided both into 'parameters' and 'alias_parameter'. "
                                f"This is not expected, as we can't know which one to keep, and which one to discard.")

            _dict_the_parameters[k] = ctx_context.obj[v]

    # We can now call the ExcelDashboarder
    _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

    try:
        _the_excel_dashboard.excelize(
            p_output_file_path=Path(file),
            dict_config=_dict_the_config,
            **_dict_the_parameters
        )
    except Exception as an_exception:
        click.secho('An exception was thrown!!!', blink=True, fg='red')
        click.secho(f"{an_exception}", fg='blue')
        raise Exception(f"") from an_exception

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import base64

from dash import dcc, callback, Output, Input, no_update

import dash_bootstrap_components as dbc

from inspect import stack
from logging import Logger, getLogger

from dash import html
from urllib.parse import unquote

from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.sheet import Sheet, SheetAsDiv
from com.enovation.toolbox.dash.widget import Widget


class NavigationAsNavbar(Widget, dbc.Navbar):

    _logger: Logger = getLogger(__name__)

    # ##################################################################################################################
    # CONSTANTS
    # ##################################################################################################################

    const_str_default_navbar_brand: str = "eNOVAtion"
    const_str_default_str_logo: str = "images/com-enovation.square.png"

    # The labels in the configuration file
    const_str__config_label__dict_properties: str = "properties"
    const_str__config_label__str_navbar_brand: str = "navbar_brand"
    const_str__config_label__url_logo: str = "logo"
    const_str__config_label__str_title: str = "title"

    # The wid for the default generic widgets
    const_str__wid__div_debug: str = "debug"
    const_str__wid__location_url: str = "url"
    const_str__wid__navbar_header: str = "header"

    """ 
    Return the navigation bar, along with the callback function to load the enov-content__as_main part
    - An image as the logo
    - A brand, which is used as well as the title for the logo
    - Menus to various widgets available.

    The produced HTML is of this kind:
    <nav id="enov-header__as_navbar" class="fixed-top navbar navbar-expand-md navbar-dark bg-dark">
        <div class="nav-item">
            <a href="https://pypi.org/project/com-enovation/">
                <div class="align-items-center row">
                    <div class="col">
                        <img height="30px" src="/assets/images/com-enovation.square.png" title="com.enovation">
                    </div>
                    <div class="col">
                        <span class="navbar-brand">eNOVAtion</span>
                    </div>
                </div>
            </a>
        </div>
        <div class="nav-item">
            <a href="#welcome" class="nav-link">welcome</a>
        </div>
        <div class="nav-item">
            <a href="#explore" class="nav-link">explore</a>
        </div>
        <div class="nav-item">
            <a class="nav-link disabled" disabled="">The default dp_grapher for date predictability module...</a>
        </div>
    </nav>        
    """

    def __init__(
            self,
            dict_sheet: dict[str, Sheet],
            dict_config: dict,
            dict_context: dict,
            str__wid__main_content: str,
            session: Session
    ):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _str_the_title: str = dict_config.get(
            self.const_str__config_label__dict_properties, {}
        ).get(self.const_str__config_label__str_title, self.const_str_default_navbar_brand)

        if self.const_str__config_label__url_logo in dict_config.get(
                self.const_str__config_label__dict_properties, {}
        ):
            _img_the_logo: html.Img = html.Img(
                src='data:image/png;base64,{}'.format(
                    base64.b64encode(
                        open(
                            dict_config[
                                self.const_str__config_label__dict_properties
                            ][
                                self.const_str__config_label__url_logo
                            ],
                            'rb'
                        ).read()
                    )
                ),
                height="30px",
                title=_str_the_title
            )
        else:
            _img_the_logo: html.Img = html.Img(
                src=session.dash_application.dash_server.get_asset_url(self.const_str_default_str_logo),
                height="30px",
                title=_str_the_title
            )

        Widget.__init__(
            self,
            dict_config=dict_config,
            dict_context=dict_context,
            session=session
        )

        dbc.Navbar.__init__(
            self,
            children=[
                # To get the url loaded
                dcc.Location(id=self.const_str__wid__location_url),

                dbc.NavItem(
                    html.A(
                        dbc.Row(
                            [
                                dbc.Col(_img_the_logo),
                                dbc.Col(
                                    dbc.NavbarBrand(dict_config.get(
                                        self.const_str__config_label__dict_properties, {}
                                    ).get(
                                        self.const_str__config_label__str_navbar_brand,
                                        self.const_str_default_navbar_brand)
                                    )
                                ),
                            ],
                            align="center",
                        ),
                        href="https://pypi.org/project/com-enovation/"
                    )
                )
            ] + [
                dbc.NavItem(dbc.NavLink(i_widget, href="#"+i_widget))
                for i_widget in dict_sheet
            ] + [
                dbc.NavItem(dbc.NavLink("The default dp_grapher for the date predictability module...", disabled=True))
            ],
            color="dark",
            dark=True,
            id=self.const_str__wid__navbar_header,
            className="fixed-top"
        )

        @callback(
            Output(str__wid__main_content, 'children'),
            Input(self.const_str__wid__location_url, 'hash'),
            prevent_initial_call=True
        )
        def _callback_navigate_and_load_content(url_hash: str) -> html.Div:
            """
            Callback function:
            - When we click in a menu in the navigation bar, it refreshes the page with an updated URL
              "https://xxx/#whatever"
            - When such URL update happens, the widget "url as location" is refreshed
            - Using this refresh, we load the area "content as main" with the new target widget.

            :param url_hash: the link to an element with a specified id within the page
            :return: the widget to load in the area "content as main"
            """
            getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

            if url_hash == '':
                return no_update

            _dict_the_sheets: dict[str, SheetAsDiv] = self.session.dash_application.dict_sheets

            if unquote(url_hash)[1:] not in _dict_the_sheets:
                raise Exception(f"Unexpected url_hash '{unquote(url_hash)}'.")

            div_the_return: html.Div = _dict_the_sheets[unquote(url_hash)[1:]]

            self.session.debug(
                debug_trace=f"Click navigation menus to update main content: {unquote(url_hash)}, returned an "
                            f"object of type '{type(div_the_return)}'"
            )

            getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

            return div_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import os
import pkgutil
from inspect import stack
from logging import Logger, getLogger

from com.enovation.toolbox.excel_dashboard.excel_dashboarder import ExcelDashboarder

_logger: Logger = getLogger(__name__)

_logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

_str_the_widgets_sub_directory: str = os.path.join(
    os.path.dirname(__file__),
    ExcelDashboarder.const_str__sub_directory_for_widgets
)
_logger.info(f"\tDirectory '{_str_the_widgets_sub_directory}' will be screened to load widget modules.")

__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages([_str_the_widgets_sub_directory]):

    __all__.append(module_name)
    _module = loader.find_module(module_name).load_module(module_name)
    globals()[module_name] = _module
    _logger.info(f"\t-- {module_name}: added")

_logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning, after having loaded "
             f"'{len(__all__)}' modules.")



from datetime import datetime, timedelta
from inspect import stack
from logging import Logger, getLogger

from enlighten import Manager, Counter
from numpy import where, ndarray, NaN
from pandas import DataFrame, MultiIndex, NaT, Series
from scipy.special import psi

from com.enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
import enlighten

from com.enovation.toolbox.predictability.bean import PredictabilityBean


class DatePredictabilityComputer:
    _logger: Logger = getLogger(__name__)

    # Columns labels for the input file that is expected from the client
    str__input__column_label__key: str = "key"
    str__input__column_label__date: str = "date"
    str__input__column_label__measure: str = "measure"
    lst__input__column_labels: list = [
        str__input__column_label__key,
        str__input__column_label__date,
        str__input__column_label__measure,
    ]

    # Columns labels for the statistics
    str__statistics__column_label__measure_count: str = "measure_count"
    str__statistics__column_label__date_first: str = "date_first"
    str__statistics__column_label__date_last: str = "date_last"
    str__statistics__column_label__measure_first: str = "measure_first"
    str__statistics__column_label__measure_last: str = "measure_last"
    str__statistics__column_label__measure_min: str = "measure_min"
    str__statistics__column_label__measure_max: str = "measure_max"
    str__statistics__column_label__predictability_last: str = "predictability_last"

    lst__output__statistics_by_key__column_labels: list = [
        str__input__column_label__key,
        str__statistics__column_label__measure_count,
        str__statistics__column_label__date_first,
        str__statistics__column_label__date_last,
        str__statistics__column_label__measure_first,
        str__statistics__column_label__measure_last,
        str__statistics__column_label__measure_min,
        str__statistics__column_label__measure_max,
        str__statistics__column_label__predictability_last
    ]

    # Columns labels for working columns
    str__output__column_label__predictability: str = "predictability"
    str__output__column_label__duration: str = "duration"

    # Default resampling frequency: 1D
    _str__default_resampling_rule: str = "1D"

    def compute_historical_date_predictability(
            self,
            df_measures: DataFrame,
            b_resample: bool = True,
            int_sliding_window: int = None
    ) -> PredictabilityBean:
        """
        Function to compute date predictability for a set of measures provided as parameter.

        The parameter df_measures should comply to the below:
        - Instance of pandas.DataFrame
        - BR_Structure_001: the measures do not have multiindex index
        - BR_Structure_002: the measures do not have multiindex columns
        - BR_Structure_003: the measures contains exactly 3 columns:
            - str__input__column_label__key: str = "key"
            - str__input__column_label__date: str = "date"
            - str__input__column_label__measure: str = "measure"
        - BR_Data_001: at least 1 record is required in order to compute a predictability
        - BR_Data_002: there should not be any null value in the dataframe
        - BR_Data_003: both columns "date" and "measure" should be dates, or objects that can be typed to dates
        - BR_Data_004: there should not be any line with duplicated "date"

        Depending on the parameter b_resample, the logic differs:
        - If b_resample == True (default):
            - The logic takes more time, but enable to graph precisely the historical trend of the date_predictability
            - The measures are resampled daily. Illustration:
                - For an input dataframe that contains 2 rows, measure 'x' for date '1-Jan' and measure 'y' for date
                  '3-Jan'
                - We resample daily, leading to dataframe that contains 3 rows, measure 'x' for dates '1-Jan' and
                  '2-Jan' and measure 'y' for date '3-Jan'
            - The date predictability is then computed as
                - SUM for_each_measure(spot_predictability)
                - With spot_predictability =
                    - 0 if measure beyond the convergence_cone
                    - otherwise (measure - last_measure)/convergence_cone
                - With convergence_cone defined by:
                    - Upper limit: date + abs(date - last_measure)
                    - Lower limit: date - abs(date - last_measure)
        - If b_resample == False:
            - The logic is more optimal than when sampling, but graphing the historical trend of the
              date_predictability is far less meaningful, and even misleading
            - The date predictability is computed as:
                - For each measure that spans from its date until the next one, we compute date_predictability as
                    - SUM for_each_day (spot_predictabilty) with spot_predictability as defined above

        Regarding parameter int_sliding_window:
        - It can only be provided when b_resample is equal to True, or an exception is thrown
        - It set the length in days of the sliding window that filters historical measures: only measures within
          this sliding window are considered, the older ones are ignored
        - By default, all historical measures are considered, even the oldest ones

        :param df_measures: a dataframe that contains the measures
        :param b_resample: whether or not we want to resample daily the measures, for later graphing
        :return: the date predictability for the measures provided as a parameter
        """
        self._logger.debug(
            f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
            f"\n\t- df_measures: '{df_measures.shape}'"
        )

        # 1. We check structure
        self._check_structure(
            df_measures=df_measures,
            lst_column_labels=self.lst__input__column_labels,
        )
        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': structure checked.")

        # 2. We check and type the data
        the_df_typed = self._check_and_cleanse_data(
            df_measures=df_measures.copy(),
            str_column_label_key=self.str__input__column_label__key,
            str_column_label_date=self.str__input__column_label__date,
            str_column_label_measure=self.str__input__column_label__measure,
        )

        # 2b. We check if a sliding window is set, that b_sample is true
        if (
            (int_sliding_window is not None)
            & (b_resample is False)
        ):
            raise Exception(f"Function '{stack()[0].filename} - {stack()[0].function}' is not properly called: "
                            f"parameter int_sliding_window is not null, equalt to '{int_sliding_window}', while"
                            f"parameter b_sample is '{b_resample}'.")

        # 3. Depending on whether or not we need to resample, we call the function to compute historical predictability
        if b_resample:
            df_the_historical_predictability: DataFrame = self._resample_and_compute_historical_date_predictability(
                df_measures=the_df_typed.copy(),
                str_column_label_key=self.str__input__column_label__key,
                str_column_label_date=self.str__input__column_label__date,
                str_column_label__measure=self.str__input__column_label__measure,
                str_column_label__predictability=self.str__output__column_label__predictability,
                str_resample_rule=self._str__default_resampling_rule,
                int_sliding_window=int_sliding_window
            )
        else:
            df_the_historical_predictability: DataFrame = \
                self._compute_historical_date_predictability_without_resampling(
                    df_measures=the_df_typed.copy(),
                    str_column_label__key=self.str__input__column_label__key,
                    str_column_label__date=self.str__input__column_label__date,
                    str_column_label__measure=self.str__input__column_label__measure,
                    str_column_label__duration=self.str__output__column_label__duration,
                    str_column_label__predictability=self.str__output__column_label__predictability,
                )

        # 4. We compute the statistics
        df_the_statistics_by_key: DataFrame = self.compute_statistics_by_key(
            df_measures=the_df_typed.copy(),
            df_historical_predictability=df_the_historical_predictability,
            str_column_label__key=self.str__input__column_label__key,
            str_column_label__date=self.str__input__column_label__date,
            str_column_label__measure=self.str__input__column_label__measure,
            str_column_label__predictability=self.str__output__column_label__predictability,
            str_column_label__measure_count=self.str__statistics__column_label__measure_count,
            str_column_label__date_first=self.str__statistics__column_label__date_first,
            str_column_label__date_last=self.str__statistics__column_label__date_last,
            str_column_label__measure_first=self.str__statistics__column_label__measure_first,
            str_column_label__measure_last=self.str__statistics__column_label__measure_last,
            str_column_label__measure_min=self.str__statistics__column_label__measure_min,
            str_column_label__measure_max=self.str__statistics__column_label__measure_max,
            str_column_label__predictability_last=self.str__statistics__column_label__predictability_last
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return PredictabilityBean(
            df_historical=df_the_historical_predictability,
            df_by_measure=df_measures,
            df_by_key=df_the_statistics_by_key,
        )

    def _check_structure(
            self,
            df_measures: DataFrame,
            lst_column_labels: list = None,
    ):
        """
        The measures provided by the client should comply to the following business rules:
            - BR_Structure_001: the measures do not have multiindex index
            - BR_Structure_002: the measures do not have multiindex columns
            - BR_Structure_003: the measures contains exactly 3 columns:
                - str__input__column_label__key: str = "key"
                - str__input__column_label__date: str = "date"
                - str__input__column_label__measure: str = "measure"

        The function:
            - Returns quietly in case the measures comply to these business rules
            - Or raises an alert in case the client's measures do not comply to any of these business rules.

        :param df_measures: the client's measures
        :param lst_column_labels: the expected columns labels
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if lst_column_labels is None:
            lst_column_labels = self.lst__input__column_labels

        # BR_Structure_001: the measures do not have multiindex index
        if isinstance(df_measures.index, MultiIndex):
            raise Exception(f"BR_Structure_001 - Provided measures have a multiindex index, which is not expected. "
                            f"Flatten the index, and retry. Multiindex names: "
                            f"'{', '.join(df_measures.index.names)}'.")

        # BR_Structure_002: the measures do not have multiindex columns
        if isinstance(df_measures.columns, MultiIndex):
            raise Exception(f"BR_Structure_002 - Provided measures have multiindexed columns, which is not expected. "
                            f"Flatten the columns, and retry. Multiindex names: "
                            f"'{', '.join(df_measures.columns.names)}'.")

        # BR_Structure_003: the measures contains exactly 3 columns, aka "key", "date" and "measure"
        lst_the_missing_columns_in_df: list = list(
            set(lst_column_labels)
            - set(df_measures.columns)
        )
        if len(lst_the_missing_columns_in_df) > 0:
            raise Exception(f"BR_Structure_003 - Provided measures are missing the following mandatory columns: "
                            f"'{', '.join(lst_the_missing_columns_in_df)}'. Review your data, and retry.")
        lst_the_extra_columns_in_df: list = list(
            set(df_measures.columns)
            - set(lst_column_labels)
        )
        if len(lst_the_extra_columns_in_df) > 0:
            raise Exception(f"BR_Structure_003 - Provided measures contains extra unexpected columns: "
                            f"'{', '.join(lst_the_missing_columns_in_df)}'. Review your data, and retry.")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _check_and_cleanse_data(
            self,
            df_measures: DataFrame,
            str_column_label_key: str = str__input__column_label__key,
            str_column_label_date: str = str__input__column_label__date,
            str_column_label_measure: str = str__input__column_label__measure,
    ) -> DataFrame:
        """
        The measures provided by the client should comply to the following business rules:
            - BR_Data_001: at least 1 record is required in order to compute a predictability
            - BR_Data_002: there should not be any null value in the dataframe
            - BR_Data_003: both columns "date" and "measure" should be dates, or objects that can be typed to dates
            - BR_Data_004: there should not be any line with duplicated "date"

        Cleansing logic:
            - LOG_Data_001: Columns "date" and "measure" are typed into dates
            - LOG_Data_002: Remove useless measures, aka same measure duplicated across dates

        The function:
            - Returns quietly in case the client's measures comply to these business rules
            - Or raises an alert in case the client's measures do not comply to any of these business rules.

        :param df_measures: the client's measures
        :param str_column_label_date: the label of the column that bears the dates for the measures
        :param str_column_label_measure: the label of the column that bears the measures
        :return: the typed client's measures
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # BR_Data_001. At least one record
        if len(df_measures.index) == 0:
            raise Exception(f"BR_Data_001 - The parameter df_measures contains '0' records, while it should contain at "
                            f"least '1'.")

        # BR_Data_002. No null value
        if df_measures.isnull().values.any():
            raise Exception(f"BR_Data_002 - The parameter df_measures contains '"
                            f"{df_measures.isnull().values.sum()}' null values, which is not expected. Make "
                            f"sure you call the function with a dataframe df_measures which does not contain any null "
                            f"value.")

        # BR_Data_003. We retype both columns "date" and "measure" into datetime:
        # LOG_Data_001: Columns "date" and "measure" are typed into dates
        try:
            df_the_return: DataFrame = PandasDataframeTyper.type(
                df_to_type=df_measures,
                dict_columns_to_type={
                    str_column_label_date: "date",
                    str_column_label_measure: "date"
                },
                b_strict=False
            )
        except Exception as an_exception:
            raise Exception(
                f"BR_Data_003 - An exception occurred while typing columns '{str_column_label_date}' and "
                f"'{str_column_label_measure}' to dates. Check the data, and retry.") from an_exception

        # BR_Data_004. duplicated records for a given "date" (excluding the "measure" column)
        df_the_working_df: DataFrame = df_measures.drop(columns=[str_column_label_measure])
        df_the_working_df = df_the_working_df[df_the_working_df.duplicated()]
        if len(df_the_working_df.index) > 0:
            the_unique_duplicated_records: DataFrame = df_the_working_df.drop_duplicates()
            the_msg: str = f"BR_Data_004 - The parameter df_measures contains '{len(df_the_working_df.index)}' " \
                           f"records duplicated, which is not expected. Make sure your remove all duplicates for " \
                           f"values:\n\n{the_unique_duplicated_records.to_string()}\n\n"
            raise Exception(the_msg)

        # LOG_Data_002: Remove useless measures, aka same measure duplicated across dates
        df_the_return = df_the_return.sort_values(
            by=[
                str_column_label_key,
                str_column_label_date
            ]
        ).reset_index(drop=True)
        df_the_return["tmp_" + str_column_label_key] = df_the_return[str_column_label_key].shift(1)
        df_the_return["tmp_" + str_column_label_measure] = df_the_return[str_column_label_measure].shift(1)

        # We override the very last measure for each key, not to drop it in case it is similar to previous one
        # df_the_return = df_the_return.loc[-1, "tmp_"+str_column_label_key] = NaN
        # df_the_return.loc[-1, "tmp_"+str_column_label_measure] = NaT
        df_the_return.loc[
            df_the_return.groupby(str_column_label_key).tail(1).index,
            ["tmp_" + str_column_label_key, "tmp_" + str_column_label_measure]
        ] = [NaN, NaT]

        df_the_return = df_the_return[
            (df_the_return[str_column_label_key] != df_the_return["tmp_" + str_column_label_key])
            | (df_the_return[str_column_label_measure] != df_the_return["tmp_" + str_column_label_measure])
            ]
        df_the_return = df_the_return.drop(columns=[
            "tmp_" + str_column_label_key,
            "tmp_" + str_column_label_measure
        ])

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        return df_the_return

    def _resample_and_compute_historical_date_predictability(
            self,
            df_measures: DataFrame,
            str_column_label_key: str = str__input__column_label__key,
            str_column_label_date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,
            str_column_label__predictability: str = str__output__column_label__predictability,
            str_resample_rule: str = _str__default_resampling_rule,
            int_sliding_window: int = None
    ) -> DataFrame:

        # 1. We resample
        the_df_resampled: DataFrame = self._resample_measures(
            df_measures=df_measures,
            str_column_label_key=str_column_label_key,
            str_column_label_date=str_column_label_date,
            str_resample_rule=str_resample_rule,
        )
        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': measures resampled.")

        # 2. We compute historical predictability
        the_df_historical_predictability: DataFrame = \
            self._compute_historical_date_predictability_for_resampled_measures(
                df_measures=the_df_resampled,
                str_column_label__key=str_column_label_key,
                str_column_label__date=str_column_label_date,
                str_column_label__measure=str_column_label__measure,
                str_column_label__predictability=str_column_label__predictability,
                int_sliding_window=int_sliding_window
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return the_df_historical_predictability

    def _resample_measures(
            self,
            df_measures: DataFrame,
            str_column_label_key: str = str__input__column_label__key,
            str_column_label_date: str = str__input__column_label__date,
            str_resample_rule: str = _str__default_resampling_rule,
    ) -> DataFrame:

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        df_measures = df_measures \
            .set_index(str_column_label_date) \
            .groupby(str_column_label_key) \
            .resample(str_resample_rule) \
            .ffill() \
            .reset_index(level=str_column_label_date) \
            .reset_index(drop=True)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return df_measures

    def _compute_historical_date_predictability_for_resampled_measures(
            self,
            df_measures: DataFrame,
            str_column_label__key: str = str__input__column_label__key,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,
            str_column_label__predictability: str = str__output__column_label__predictability,
            int_sliding_window: int = None
    ) -> DataFrame:
        """
        A function to compute a date predictability across history.

        :param df_measures: measures for a given key
        :param int_sliding_window: the length in days of the sliding window to compute predictability. Old measures
          that falls outside this sliding window are ignored. By default, this filter is ignored, and all
          historical data, ever the oldest ones, are considered.
        :return: the historical predictability that would have been computed as of the last date.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_measures[str_column_label__predictability] = -1

        # Setup progress bar
        obj_the_mgr: Manager = enlighten.get_manager()
        obj_the_counter: Counter = obj_the_mgr.counter(
            total=len(df_measures.index),
            desc='Progress:',
            unit='measures'
        )

        # For each line in the dataframe
        for t_i_row in df_measures.itertuples():
            # We compute the predictability:
            # - For the given key
            # - For all the previous records at the date

            # We filter the past measures for a given key
            df_the_filtered_measures: DataFrame = df_measures[
                (df_measures[str_column_label__key] == getattr(t_i_row, str_column_label__key))
                & (df_measures[str_column_label__date] <= getattr(t_i_row, str_column_label__date))
            ]

            # If the sliding window is defined
            if int_sliding_window is not None:
                df_the_filtered_measures = df_the_filtered_measures[
                    df_measures[str_column_label__date] >= (
                        getattr(t_i_row, str_column_label__date) -
                        timedelta(days=int_sliding_window)
                    )
                ]

            # Eventually, we call the function to compute the data predictability for the key and date
            f_i_predictability = self._compute_date_predictability_for_a_given_key_and_resampled_measures(
                df_measures=df_the_filtered_measures,
                str_column_label__date=str_column_label__date,
                str_column_label__measure=str_column_label__measure,
            )

            # We update the measures with these historical predictabilities
            df_measures.loc[getattr(t_i_row, 'Index'), str_column_label__predictability] = f_i_predictability

            # We update the progress bar
            obj_the_counter.update()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return df_measures

    def _compute_date_predictability_for_a_given_key_and_resampled_measures(
            self,
            df_measures: DataFrame,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure
    ) -> float:
        # Too many logs:
        # self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        dt_the_last_date: datetime = df_measures[self.str__input__column_label__date].max()

        dt_the_last_measure: datetime = \
            df_measures[
                df_measures[self.str__input__column_label__date] == dt_the_last_date
                ][
                self.str__input__column_label__measure
            ].iloc[0]

        a_measures_predictabilities: ndarray = where(

            # If date is at last measure
            df_measures[str_column_label__date] == dt_the_last_measure,

            where(

                # If measure is spot on
                df_measures[str_column_label__measure] == dt_the_last_measure,
                1,
                0
            ),

            where(

                # If outside the cone
                abs((df_measures[str_column_label__measure] - dt_the_last_measure).dt.days)
                >= abs((df_measures[str_column_label__date] - dt_the_last_measure).dt.days),
                0,
                1
                - abs((df_measures[str_column_label__measure] - dt_the_last_measure).dt.days)
                / abs((df_measures[str_column_label__date] - dt_the_last_measure).dt.days)
            )
        )

        # The last measure should not contribute to the aggregated predictability
        # Therefore, we remove the last record from the ndarray (only if more than 1 records)
        if len(a_measures_predictabilities) > 1:
            a_measures_predictabilities = a_measures_predictabilities[:-1]

        # Eventually, we compute the aggregated predictability
        f_the_return: float = \
            a_measures_predictabilities.sum() \
            / len(a_measures_predictabilities)

        # Too many logs:
        # self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return f_the_return

    def _compute_historical_date_predictability_without_resampling(
            self,
            df_measures: DataFrame,
            str_column_label__key: str = str__input__column_label__key,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,
            str_column_label__predictability: str = str__output__column_label__predictability,
            str_column_label__duration: str = str__output__column_label__duration,
    ) -> DataFrame:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        the_df_enriched_measures: DataFrame = df_measures

        # Setup progress bar
        obj_the_mgr: Manager = enlighten.get_manager()
        obj_the_counter: Counter = obj_the_mgr.counter(
            total=len(df_measures.index),
            desc='Progress:',
            unit='measures'
        )

        # We instantiate the column to bear the predictability
        the_df_enriched_measures[str_column_label__predictability] = -1

        # For each line in the dataframe
        for t_i_row in df_measures.itertuples():
            # We compute the predictability:
            # - For the given key
            # - For all the previous records at the date
            f_i_predictability = self._compute_date_predictability_for_a_given_key_not_resampled(
                df_measures=the_df_enriched_measures[
                    (the_df_enriched_measures[str_column_label__key] == getattr(t_i_row, str_column_label__key))
                    & (the_df_enriched_measures[str_column_label__date] <= getattr(t_i_row, str_column_label__date))
                    ],
                str_column_label__date=str_column_label__date,
                str_column_label__measure=str_column_label__measure,
                str_column_label__duration=str_column_label__duration,
            )

            # We update the measures with these historical predictabilities
            df_measures.loc[getattr(t_i_row, 'Index'), str_column_label__predictability] = f_i_predictability

            # We update the progress bar
            obj_the_counter.update()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return df_measures

    def _compute_date_predictability_for_a_given_key_not_resampled(
            self,
            df_measures: DataFrame,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,
            str_column_label__duration: str = str__output__column_label__duration,
    ) -> float:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        the_df_enriched_measures: DataFrame = df_measures.copy()

        # We duplicate column "date", that we shift by one
        # Assumption: values already sorted in function _check_and_cleanse_data
        the_df_enriched_measures["date_end"] = the_df_enriched_measures[str_column_label__date].shift(-1)
        the_df_enriched_measures.loc[the_df_enriched_measures.index[-1], "date_end"] = \
            the_df_enriched_measures.loc[the_df_enriched_measures.index[-1],
                                         DatePredictabilityComputer.str__input__column_label__date]

        # We compute the duration for each measure
        the_df_enriched_measures[str_column_label__duration] = (
                the_df_enriched_measures["date_end"] - the_df_enriched_measures[str_column_label__date]).dt.days

        dt_the_last_date: datetime = df_measures[self.str__input__column_label__date].max()
        dt_the_last_measure: datetime = df_measures[
            df_measures[self.str__input__column_label__date] == dt_the_last_date
            ][
            self.str__input__column_label__measure
        ].iloc[0]

        def compute_magic_number(s_row: Series) -> float:
            dt_the_date_start: datetime = s_row[str_column_label__date]
            dt_the_date_end: datetime = s_row["date_end"]
            dt_the_measure: datetime = s_row[str_column_label__measure]
            int_the_duration: int = s_row[str_column_label__duration]

            # For the very last measure, the closing one,
            if int_the_duration == 0:
                return 0.0

            # If we are spot on
            if dt_the_measure == dt_the_last_measure:
                return int_the_duration

            # If date_start is before last measure, and within convergence cone
            if (dt_the_date_start < dt_the_last_measure) & (
                    abs((dt_the_measure - dt_the_last_measure).days)
                    < abs((dt_the_last_measure - dt_the_date_start).days)
            ):
                dt_the_effective_date_end: datetime = min(
                    dt_the_date_end,
                    dt_the_last_measure-timedelta(days=abs((dt_the_measure - dt_the_last_measure).days))
                )
                # int_the_cone_start: int = abs((dt_the_last_measure-dt_the_date_start).days)
                int_the_cone_effective_end: int = abs((dt_the_last_measure-dt_the_effective_date_end).days)+1

                # We prepare the operands
                int_b: int = int_the_cone_effective_end
                int_a: int = int_the_cone_effective_end - abs((dt_the_last_measure-dt_the_measure).days)
                int_m: int = (dt_the_effective_date_end-dt_the_date_start).days

                f_the_predictability: float = ((int_b-int_a)*(psi(int_b)-psi(int_b+int_m))+int_m)
            else:
                f_the_predictability: float = 0

            # If date_end is after last measure, and within convergence cone
            if (dt_the_date_end > dt_the_last_measure) & (
                    abs((dt_the_measure - dt_the_last_measure).days)
                    < abs((dt_the_last_measure - dt_the_date_end).days)
            ):
                dt_the_effective_date_start: datetime = max(
                    dt_the_date_start,
                    dt_the_last_measure+timedelta(days=abs((dt_the_measure - dt_the_last_measure).days))
                )
                # int_the_cone_end: int = abs((dt_the_last_measure-dt_the_date_end).days)-1
                int_the_cone_effective_start: int = abs((dt_the_last_measure-dt_the_effective_date_start).days)

                # We prepare the operands
                int_b: int = int_the_cone_effective_start
                int_a: int = int_the_cone_effective_start - abs((dt_the_last_measure-dt_the_measure).days)
                int_m: int = (dt_the_date_end-dt_the_effective_date_start).days

                f_the_predictability += ((int_b-int_a)*(psi(int_b)-psi(int_b+int_m))+int_m)
            else:
                pass

            return f_the_predictability

        if len(df_measures.index) == 1:
            f_the_return: float = 1.0

        else:
            the_s_historical_predictability: Series = the_df_enriched_measures.apply(compute_magic_number, axis=1)

            # Eventually, we compute the aggregated predictability
            f_the_return: float = \
                the_s_historical_predictability.sum() / \
                the_df_enriched_measures[str_column_label__duration].sum()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return f_the_return

    def compute_statistics_by_key(
            self,
            df_measures: DataFrame,
            df_historical_predictability: DataFrame,

            str_column_label__key: str = str__input__column_label__key,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,

            str_column_label__predictability: str = str__output__column_label__predictability,

            str_column_label__measure_count: str = str__statistics__column_label__measure_count,
            str_column_label__date_first: str = str__statistics__column_label__date_first,
            str_column_label__date_last: str = str__statistics__column_label__date_last,
            str_column_label__measure_first: str = str__statistics__column_label__measure_first,
            str_column_label__measure_last: str = str__statistics__column_label__measure_last,
            str_column_label__measure_min: str = str__statistics__column_label__measure_min,
            str_column_label__measure_max: str = str__statistics__column_label__measure_max,
            str_column_label__predictability_last: str = str__statistics__column_label__predictability_last
    ) -> DataFrame:
        """
        todo.

        :param str_column_label__predictability_last:
        :param str_column_label__measure_max:
        :param str_column_label__measure_min:
        :param str_column_label__measure_last:
        :param str_column_label__measure_first:
        :param str_column_label__date_last:
        :param str_column_label__date_first:
        :param str_column_label__measure_count:
        :param str_column_label__predictability:
        :param str_column_label__measure:
        :param str_column_label__date:
        :param str_column_label__key:
        :param df_measures:
        :param df_historical_predictability: measures for a given key
        :return: the historical predictability that would have been computed as of the last date.
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. We compute global statistics for each predictability.V2.V2 key:
        #       - date_first
        #       - date_last
        #       - measure_min
        #       - measure_max
        df_the_stats_by_key: DataFrame = df_measures.groupby(str_column_label__key).agg({
            str_column_label__key: "count",
            str_column_label__date: ["min", "max"],
            str_column_label__measure: ["min", "max"],
        }).reset_index()
        df_the_stats_by_key.columns = [
            str_column_label__key,
            str_column_label__measure_count,
            str_column_label__date_first,
            str_column_label__date_last,
            # str_column_label__measure_first,
            # str_column_label__measure_last,
            str_column_label__measure_min,
            str_column_label__measure_max,
            # str__output__column_label__predictability
        ]

        # 2. We compute the 2 residual statistics from the measures depending on date first and last:
        #    - measure_first
        #    - measure_last
        df_the_stats_by_key = df_the_stats_by_key.merge(
            right=df_measures.rename(columns={
                str_column_label__date: str_column_label__date_first,
                str_column_label__measure: str_column_label__measure_first
            }),
            on=[str_column_label__key, str_column_label__date_first],
            how="left"
        )
        df_the_stats_by_key = df_the_stats_by_key.merge(
            right=df_measures.rename(columns={
                str_column_label__date: str_column_label__date_last,
                str_column_label__measure: str_column_label__measure_last
            }),
            on=[str_column_label__key, str_column_label__date_last],
            how="left"
        )

        # 3. We compute the statistic based on the historical predictabilities
        #    - predictability_last
        df_the_stats_by_key = df_the_stats_by_key.merge(
            right=df_historical_predictability[[
                str_column_label__key,
                str_column_label__date,
                str_column_label__predictability,
            ]].rename(columns={
                str_column_label__date: str_column_label__date_last,
                str_column_label__predictability: str_column_label__predictability_last
            }),
            on=[str_column_label__key, str_column_label__date_last],
            how="left"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return df_the_stats_by_key



from dash.html import Div

from com.enovation.toolbox.dash.session import Session


class Widget:

    @property
    def session(self):
        return self._session

    @property
    def dict_config(self) -> dict:
        return self._dict_config

    @property
    def dict_context(self) -> dict:
        return self._dict_context

    def __init__(
            self,
            dict_config: dict,
            dict_context: dict,
            session: Session
    ):

        self._dict_config: dict = dict_config
        self._dict_context: dict = dict_context
        self._session: Session = session


class WidgetAsDiv(Widget, Div):

    def __init__(
            self,
            dict_config: dict,
            dict_context: dict,
            session: Session,
            *args, **kwargs
    ):
        # We initialize the widget
        Widget.__init__(
            self,
            dict_config=dict_config,
            dict_context=dict_context,
            session=session
        )

        # We initialize the Div object...
        Div.__init__(
            self,
            *args,
            **kwargs
        )



import logging
from inspect import stack

import click
from click import BadParameter

from com.enovation.toolbox.dash.application import DashApplication

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('ds-dash')
@click.pass_context
@click.argument('configuration')
def ds_dash(ctx_context, configuration):
    """
    Launch dash application according to the CONFIGURATION and consuming DATA.

    CONFIGURATION is a json file that list the widgets to load, along with all the relevant parameters, such as the
    data to get from the context
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if configuration not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot find argument '{configuration}' in context, as the dictionary configuration of the dash "
                    f"application.",
            param_hint=f"Ensure you effectively load in the context a dictionary configuration for the dash "
                       f"application to know what widgets to load, along with their parameters.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[configuration], dict):
        raise BadParameter(
            message=f"Argument '{configuration}' is of type '{type(ctx_context.obj[configuration])}' while "
                    f"'dict' is expected.",
            param_hint=f"Ensure you effectively provide an argument '{configuration}' of type 'dict'.",
            ctx=ctx_context
        )

    _obj_the_dash_app: DashApplication = DashApplication().launch(
        dict_config=ctx_context.obj[configuration],
        dict_context=ctx_context.obj
    )

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger

from dash import callback, Output, Input
from dash.html import P

import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_wid as wid
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_session import Session

_logger: Logger = getLogger(__name__)


@callback(
    Output(wid.str_the_wid__debug_as_samp, 'children'),
    Input(wid.str_the_wid__debug_as_interval, 'n_intervals'),
)
def _callback_debug(i_interval_n: int) -> list[P]:
    """
    Callback function to update the debug
    :param i_interval_n: number of seconds since the application started
    :return:
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameter i_interval_n = "
                  f"{i_interval_n}, and returning.")

    return Session.get_debug()



from inspect import stack
from logging import Logger, getLogger

from dash import html

from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.widget import Widget


class FooterAsFooter(Widget, html.Footer):
    """

    Return the footer such as:
    <footer class="footer" id="enov-footer__as_footer">
        <div class="container">
            <span class="text-muted">Copyright © 2022 Jean-Sébastien Guillard. All rights reserved.</span>
        </div>
    </footer>
    """
    _logger: Logger = getLogger(__name__)

    # The widget IDs to be used
    # Naming convention:
    # - should start by "const_str__wid__"
    # - as the application will screen all the widget ids, across widgets, to ensure there is no ambiguity, aka same
    #   widget id across different widgets
    const_str__wid__footer_footer: str = "footer"

    def __init__(
            self,
            dict_config: dict,
            dict_context: dict,
            session: Session
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        Widget.__init__(
            self,
            dict_config=dict_config,
            dict_context=dict_context,
            session=session
        )

        html.Footer.__init__(
            self,

            children=[
                html.Div(
                    [
                        html.Span(
                            [
                                "Copyright © 2022 Jean-Sébastien Guillard. All rights reserved."
                            ],
                            className="text-muted"
                        )
                    ],
                    className="container",
                ),
            ],
            className="footer",
            id=self.const_str__wid__footer_footer,

        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import inspect
import logging

from numpy import where
from pandas import DataFrame


class PandasDataframeSampler:
    _logger: logging.Logger = logging.getLogger(__name__)

    @staticmethod
    def compress(
            df_measures: DataFrame,
            lst_key_columns: list = None,
            lst_value_columns: list = None,
            lst_ordering_columns: list = None,
            b_keep_last: bool = False,
    ) -> DataFrame:
        """
        The function screens the rows within each universe one by one, and it drops the rows which are equal to the
        previous one.

        Three list of columns are provided:
        - lst_key_columns are used to define the universes. Each universe is processed independently from the others. In
          other words, calling the function universe by universe would return the same result.
        - lst_value_columns are reconciled, and if all values are equal, then 2 rows are deemed identical. In case at
          least one value is different, then rows are considered NOT identical.
        - lst_ordering_columns are used to sort the values in the dataframe before processing the lines one by one. If
          the parameter is null, the function assume the rows are already sorted.

        If argument b_keep_last is set to True, the very last record of each key is kept, even if it is a duplicate
        with a previous row.

        BR_001, dataframe expected not to contain several columns with the same label
        BR_002, all columns from lst_key_columns expected to be in the dataframe
        BR_003, all columns from lst_value_columns expected to be in the dataframe
        BR_004, all columns from lst_ordering_columns expected to be in the dataframe
        BR_005, all columns from dataframe should be listed into one of the list: key, value or ordering
        BR_006, dataframe expected not to contain any duplicate across the ordering columns (within the same universe)

        :param df_measures: the dataframe from which duplicates will be dropped
        :param lst_key_columns: lst_key_columns are used to define the universes. Each universe is processed
            independently from the others
        :param lst_value_columns: lst_value_columns are reconciled, and if all values are equal, then 2 rows are deemed
            identical. In case at least one value is different, then rows are considered NOT identical
        :param lst_ordering_columns: lst_ordering_columns are used to sort the values in the dataframe before
            processing the lines one by one. If the parameter is null, the function assume the rows are already sorted.
        :param b_keep_last: whether or not the last row of each keys is to be kept even if duplicate
        """

        PandasDataframeSampler._logger.debug(
            f"Function '{inspect.stack()[0].filename} - {inspect.stack()[0].function}' is called.")

        # We duplicate the data, not to alter the source...
        df_measures = df_measures.copy()

        # Not to have conflicts with working columns that will be added in the below logic...
        df_measures.columns = ["_" + i_col for i_col in df_measures.columns]

        # ############################################################################################################ #
        # ## First step: we prepare the various lists, as we authorize null values...
        # ## - If key columns are null, we consider an empty list
        # ## - If ordering columns are null, we consider an empty list
        # ##   - Otherwise, we ensure there is not duplicate with the key columns
        # ## - If value columns are null, we consider all the columns from the dataframe, minus the key and ordering
        # ##   columns
        # ##   - Otherwise, we ensure there is not duplicate with the key columns
        # ############################################################################################################ #

        # If no key columns are provided
        if lst_key_columns is None:
            lst_key_columns = []
        else:
            # Not to have conflicts with working columns that will be added in the below logic...
            lst_key_columns = ["_" + i_key for i_key in lst_key_columns]

        # If no ordering columns are provided
        if lst_ordering_columns is None:
            lst_ordering_columns = []
        # Particular case: to avoid duplicates, we make sure no columns are both in key and ordering columns
        else:
            # Not to have conflicts with working columns that will be added in the below logic...
            lst_ordering_columns = ["_" + i_key for i_key in lst_ordering_columns]

            # The following is buggy, as using set could impact the order in the list, thus impacting the ordering logic
            # lst_ordering_columns = list(set(lst_ordering_columns) - set(lst_key_columns))
            # Replaced by
            for i_col in lst_key_columns:
                try:
                    lst_ordering_columns.remove(i_col)
                except ValueError:
                    pass

        # If no value columns are provided, we consider all the columns from the dataframe, minus the keys and orderings
        if lst_value_columns is None:
            lst_value_columns = list(set(df_measures.columns)-set(lst_key_columns)-set(lst_ordering_columns))
            PandasDataframeSampler._logger.info(
                f"The list of value columns was None:"
                f"\n\t- We took all the columns: '{', '.join(df_measures.columns)}'"
                f"\n\t- We removed the keys: '{', '.join(lst_key_columns)}'"
                f"\n\t- We removed the ordering columns: '{', '.join(lst_ordering_columns)}'"
                f"\n\t--> We eventually returned: '{', '.join(lst_value_columns)}'.")
        # Particular case: to avoid duplicates, we make sure no columns are both in key, ordering and value columns
        else:
            # Not to have conflicts with working columns that will be added in the below logic...
            lst_value_columns = ["_" + i_key for i_key in lst_value_columns]

            lst_value_columns = list(set(lst_value_columns) - set(lst_ordering_columns) - set(lst_key_columns))

        # ############################################################################################################ #
        # ## Second step: we implement the business rules
        # ############################################################################################################ #

        # BR_001, dataframe expected not to contain several columns with the same label
        lst_all_columns: list = df_measures.columns
        lst_unique_columns: list = list(set(lst_all_columns))
        lst_duplicated_columns: list = list(set(lst_all_columns)-set(lst_unique_columns))
        if len(lst_duplicated_columns) > 0:
            raise Exception(f"BR_001, dataframe provided as a parameter does contain several columns labelled "
                            f"'{', '.join(lst_duplicated_columns)}' (discard the '_' ahead of the column labels).")

        # BR_002, all columns from lst_key_columns expected to be in the dataframe
        lst_missing_key_columns: list = list(set(lst_key_columns) - set(df_measures.columns))
        if len(lst_missing_key_columns) > 0:
            raise Exception(f"BR_002, dataframe provided as a parameter (discard the '_' ahead of the column labels)"
                            f"\n\t- does not contain key columns '{', '.join(lst_missing_key_columns)}'"
                            f"\n\t- but does only contain '{', '.join(df_measures.columns)}'")

        # BR_003, all columns from lst_value_columns expected to be in the dataframe
        lst_missing_value_columns: list = list(set(lst_value_columns) - set(df_measures.columns))
        if len(lst_missing_value_columns) > 0:
            raise Exception(f"BR_003, dataframe provided as a parameter"
                            f"\n\t- does not contain value columns '{', '.join(lst_missing_value_columns)}'"
                            f"\n\t- but does only contain '{', '.join(df_measures.columns)}' (discard the '_' ahead of "
                            f"the column labels)")

        # BR_004, all columns from lst_ordering_columns expected to be in the dataframe
        if len(lst_ordering_columns) > 0:
            lst_missing_ordering_columns: list = list(set(lst_ordering_columns) - set(df_measures.columns))
            if len(lst_missing_value_columns) > 0:
                raise Exception(f"BR_004, dataframe provided as a parameter"
                                f"\n\t- does not contain value columns '{', '.join(lst_missing_ordering_columns)}'"
                                f"\n\t- but does only contain '{', '.join(df_measures.columns)}' (discard the '_' "
                                f"ahead of the column labels)")

        # BR_005, all columns from dataframe should be listed into one of the list: key, value or ordering
        lst_all_columns: list = lst_key_columns + lst_value_columns
        if len(lst_ordering_columns) > 0:
            lst_all_columns = lst_all_columns + lst_ordering_columns
        lst_all_columns = list(dict.fromkeys(lst_all_columns))  # To remove any duplicates
        lst_missing_columns: list = list(set(df_measures.columns)-set(lst_all_columns))
        if len(lst_missing_columns) > 0:
            raise Exception(f"BR_005, dataframe has columns not typed among 'key', 'value', or 'ordering'. Ensure "
                            f"to effectively type these columns when calling the function: "
                            f"'{', '. join(lst_missing_columns)}' (discard the '_' ahead of the column labels).")

        # BR_006, dataframe expected not to contain any duplicate across the ordering columns (within the same universe)
        if len(lst_ordering_columns) > 0:
            df_keys_ordering_only: DataFrame = \
                df_measures[lst_key_columns+lst_ordering_columns] \
                .groupby(by=lst_key_columns+lst_ordering_columns) \
                .size().reset_index()
            df_keys_ordering_only.columns = lst_key_columns+lst_ordering_columns+["count"]
            if len(df_keys_ordering_only[df_keys_ordering_only["count"] > 1].index) > 0:
                raise Exception(f"BR_006, dataframe provided contains duplicates among key and ordering columns\n"
                                f"{df_keys_ordering_only[df_keys_ordering_only['count'] > 1].to_string()}")

        # ############################################################################################################ #
        # ## Third step: we sort values
        # ############################################################################################################ #

        # If not yet ordered, we sort values in the dataframe
        if len(lst_ordering_columns) > 0:
            df_measures = df_measures.sort_values(by=lst_ordering_columns)

        # ############################################################################################################ #
        # ## Fourth step: we process the dataframe
        # ############################################################################################################ #

        # We now split the dataframe by the keys into groups that are processed one by one
        if len(lst_key_columns) > 0:

            # JSG as of 22-Jan-2023: FutureWarning ---------------------------------------------------------------------
            # FutureWarning: Not prepending group keys to the result index of transform-like apply. In the future, the
            # group keys will be included in the index, regardless of whether the applied function returns a
            # like-indexed object.
            # To preserve the previous behavior, use
            # 	>>> .groupby(..., group_keys=False)
            # To adopt the future behavior and silence this warning, use
            # 	>>> .groupby(..., group_keys=True)
            # FROM -----------------------------------------------------------------------------------------------------
            # df_the_return: DataFrame = \
            #     df_measures.groupby(by=lst_key_columns).apply(
            #         func=PandasDataframeSampler.little_magic,
            #         lst_value_columns=lst_value_columns,
            #         b_keep_last=b_keep_last
            #     )
            # TO -------------------------------------------------------------------------------------------------------
            df_the_return: DataFrame = \
                df_measures.groupby(by=lst_key_columns, group_keys=False).apply(
                    func=PandasDataframeSampler.little_magic,
                    lst_value_columns=lst_value_columns,
                    b_keep_last=b_keep_last
                )
            # END ------------------------------------------------------------------------------------------------------

        else:
            df_the_return: DataFrame = PandasDataframeSampler.little_magic(
                df_group=df_measures,
                lst_value_columns=lst_value_columns,
                b_keep_last=b_keep_last
            )

        # ############################################################################################################ #
        # ## Fifth step: eventually, we prepare the return value
        # ## - We remove all temporary and working columns that were added in the "little_magic"
        # ## - We remove the "_" we added at the very beginning
        # ## - We remove the multiindex
        # ############################################################################################################ #
        df_the_return = df_the_return[lst_all_columns]
        df_the_return.columns = [i_col[1:] for i_col in df_the_return.columns]
        df_the_return = df_the_return.reset_index(drop=True)

        PandasDataframeSampler._logger.debug(
            f"Function '{inspect.stack()[0].filename} - {inspect.stack()[0].function}' is returning.")

        return df_the_return

    @staticmethod
    def little_magic(
            df_group: DataFrame,
            lst_value_columns: list,
            b_keep_last: bool,
    ) -> DataFrame:

        df_group = df_group.copy()

        # We duplicate, and shift, the value columns that we want to compare across rows
        # And we compare values
        for i_col in lst_value_columns:

            df_group["tmp_" + i_col] = df_group[i_col].shift(1)
            df_group["equal_" + i_col] = where(
                (df_group["tmp_" + i_col].isnull() & df_group[i_col].isnull())
                | (
                        ~df_group["tmp_" + i_col].isnull() & ~df_group[i_col].isnull()
                        & (df_group["tmp_" + i_col] == df_group[i_col])
                ),
                True,
                False
            )

        # We compute for each line whether or not the row is similar to the previous one
        df_group["equal"] = df_group[["equal_" + i_col for i_col in lst_value_columns]].all(axis=1)

        # The first line should not be dropped
        df_group.loc[df_group.head(1).index, "equal"] = False

        # If b_keep_last, the last line should not be dropped neither
        if b_keep_last:
            df_group.loc[df_group.tail(1).index, "equal"] = False

        # Eventually, we drop duplicated rows
        df_group = df_group[~df_group["equal"]]

        return df_group



from inspect import stack
from logging import Logger, getLogger

from dash import html, dcc

import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_wid as wid


class TabDebug(html.Div):
    _logger: Logger = getLogger(__name__)

    """
    Return the welcome widget such as:
    <div class="jumbotron">
        <h1>Date Predictability</h1>
        <p class="lead">
            'Predictability' is about accurately predicting early enough the project outcomes (incl. beyond own
            roles and responsibilities), to enable timely corrective actions that increase the likelihood of
            achieving targets and reducing outcome variance.
        </p>
        <hr class="my-2">
        <p>Explanation of how it works here...</p>
    </div>
    """

    def __init__(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # ##############################################################################################################
        # Callbacks should be imported here... Otherwise, we face "unknown id" as callbacks get registered even though
        # the widgets are not used...
        # noinspection PyUnresolvedReferences
        import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_debug.tab_debug_callbacks
        # ##############################################################################################################

        super().__init__(
            children=[
                html.Div(
                    children="Debug",
                    className="bg-secondary",
                    style={
                        "margin-left": "-0.75rem",
                        "margin-right": "-0.75rem",
                        "padding-left": "0.75rem",
                    }
                ),
                html.Div(
                    children=[
                        html.Samp(
                            children=[],
                            style={"font-size": "50%"},
                            id=wid.str_the_wid__debug_as_samp,
                        )
                    ]
                ),
                dcc.Interval(
                    id=wid.str_the_wid__debug_as_interval,
                    interval=1000,  # in milliseconds
                    n_intervals=0
                )
            ],
            id=wid.str_the_wid__debug_as_div,
            className="container border bg-light",
            style={"margin-top": "1rem"}
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack
from logging import Logger, getLogger
from dash import html

from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.widget import WidgetAsDiv


class WelcomeAsDiv(WidgetAsDiv):
    """
    Return the welcome widget such as:
    <div class="jumbotron">
        <h1>Welcome Dude!</h1>
        <p class="lead">
            Welcome to the com.enovation application, powered by Dash and Plotly.
        </p>
        <hr class="my-2">
        <p>Explanation of how it works here...</p>
    </div>
    """
    _logger: Logger = getLogger(__name__)

    def __init__(
            self,
            dict_config: dict,
            dict_context: dict,
            session: Session
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
        super().__init__(
            dict_config=dict_config,
            dict_context=dict_context,
            session=session,
            children=[
                html.H1("Welcome Dude!"),
                html.P(
                    "Welcome to the com.enovation application, powered by Dash and Plotly.",
                    className="lead",
                ),
                html.Hr(className="my-2"),
                html.P(
                    "Explanation of how it works here..."
                )
            ],
            className="jumbotron",
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import os
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path

from dash import Dash
import dash_bootstrap_components as dbc

from com.enovation.toolbox.dash.layout.layout.layout__as_div import LayoutAsDiv
from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.sheet import SheetAsDiv
from com.enovation.toolbox.dash.widget import Widget
from com.enovation.toolbox.dash.widgets.bubbles.bubbles__as_div import BubblesAsDiv
from typing import Type, Optional


class DashApplication:
    """
    The master dash application, that can be called from command lines.

    The "__init__" function prepares the framework:
    - TODO: Consolidate the list of all widgets: the default ones from com.enovation, and the ones implemented outside
    - TODO: Instantiate the ConfigChecker: used whenever the application is launched to check the application and
      widgets configurations are correct

    Then, the function "launch" launches the application:
    - TODO: Instantiate the sheets containing one or more widget(s)
    - TODO: Instantiate the dash application, and launch it

    The application is composed of 3 parts:
    - The header navbar, that contains menus to navigate across the sheets
    - The content, which displays a sheet (which itself contains widgets)
    - The footer, which contains some additional information.

    Application is configured as a json:
        {
            "properties": {
                ... the dash application properties ...
                "str_url_brand_logo": ... the url to the logo ...
                "str_brand": ... the title/ brand name that is used as the title to the logo and in the navbar ...
                "str_url_brand_site": ... the url to the brand website ...
            },
            "sheets": {
                "a name for the sheet": {
                    "properties": {
                        ... the sheet properties ...
                    },
                    "widgets": [
                        {
                            "id": "the widget id (aka "$id" from the json schema) to instantiate",
                            "properties": {
                                ... the widget properties ...
                            }
                        }
                    ]
                }
            }
        }

    """
    _logger: Logger = getLogger(__name__)

    # ##################################################################################################################
    # CONSTANTS
    # ##################################################################################################################

    # Link to the directory "assets", where web site's assets are stored (css, pictures, etc)
    const_str_relative_path_to_assets: str = '../../../../../assets/'

    # The labels in the configuration file
    const_str__config_label__dict_properties: str = "properties"
    const_str__config_label__dict_sheets: str = "sheets"
    const_str__config_label__list_widgets: str = "widgets"

    const_str__config_label__str_url_brand_logo: str = "str_url_brand_logo"
    const_str__config_label__str_brand: str = "str_brand"
    const_str__config_label__str_url_brand_site: str = "str_url_brand_site"

    const_str__config_label__str_widget_id: str = "id"

    # The wid for the default generic widgets
    const_str__wid__navbar_navbar: str = "navbar"
    const_str__wid__div_debug: str = "debug"
    const_str__wid__footer_footer: str = "footer"

    # ##################################################################################################################
    # INSTANCE PROPERTIES
    # ##################################################################################################################

    @property  # The instance of the Dash server
    def dash_server(self) -> Dash:
        return self._dash_server

    @property  # Whether in the layout, we trace debug messages
    def b_debug(self) -> bool:
        return self._b_debug

    # This is the repository of all the possible widgets:
    # - The ones proposed by com.enovation packaging
    # - Along with the custom ones that are provided while instantiating the dash application.
    # It is structured as a dictionary:
    # - Key: the widget id (aka "$id" from the json schema)
    # - Value: the widget class
    @property
    def dict_widgets_implementations_repository(self) -> dict[str, Type[Widget]]:
        return self._dict_widgets_implementations_repository

    @property
    def dict_config(self) -> dict:
        return self._dict_config

    @property
    def dict_context(self) -> dict:
        return self._dict_context

    @property
    def dict_sheets(self) -> dict[str, SheetAsDiv]:
        return self._dict_sheets

    @property
    def session(self) -> Session:
        return self._session

    def __init__(
            self,
            lst_custom_widgets_implementations: list[Type[Widget]] = None
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We check that the directory "assets" exists...
        if not Path(__file__, "../", self.const_str_relative_path_to_assets).resolve().is_dir():
            raise Exception(
                f'When trying to instantiate the Dash application, we could not find the directory assets":'
                f'\n\t- From current file path "{Path(__file__)}"'
                f'\n\t- Adding relative path "{self.const_str_relative_path_to_assets}"'
                f'\n\t- Leading to non existing directory '
                f'"{Path(__file__, "../", self.const_str_relative_path_to_assets).resolve()}'
            )

        # We initialize the list of widgets
        # - We screen the sub-directory "widgets" to identify the default widgets coming from com-enovation
        # - We consolidate with the custom widgets which are provided by consumers.
        self._dict_widgets_implementations_repository: dict[str, Type[Widget]] = \
            self._init_widgets_implementations_repository(
                lst_custom_widgets_implementations=lst_custom_widgets_implementations
            )

        # We declare the other instance properties (initialized in the function "launch")
        self._b_debug: Optional[bool] = None
        self._dash_server: Optional[Dash] = None
        self._dict_config: Optional[dict] = None
        self._dict_context: Optional[dict] = None
        self._dict_sheets: Optional[dict[str, SheetAsDiv]] = None
        self._session: Optional[Session] = None

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def _init_widgets_implementations_repository(
            self,
            lst_custom_widgets_implementations: list[Type[Widget]]
    ) -> dict[str, Type[Widget]]:
        """
        TODO: to be enriched:
        - We screen the sub-directory "widgets" to identify the default widgets coming from com-enovation
        - We consolidate with the custom widgets which are provided by consumers.

        :param lst_custom_widgets_implementations:
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        dict_the_return: dict[str, Type[Widget]] = {}

        if lst_custom_widgets_implementations is None:
            self._logger.info(f"No lst_custom_widgets_implementations provided...")

        else:
            raise Exception(f"Not yet implemented...")

        dict_the_return["https://enovation.com/dash_application/bubbles"] = BubblesAsDiv

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return dict_the_return

    def launch(
            self,
            dict_config: dict,
            dict_context: dict,
            b_debug: bool = False
    ):
        """
        Main function call to instantiate and launch dash application:
        - The layout is prepared:
          - The Navbar, the footer and the main content
            - If debug, the main content displays the debug widget
          - The various widgets, according to their configurations

        Exception:
        - When there is already a Dash app instantiated
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        if self.dash_server is not None:
            raise Exception(f"The property dash_server is not null, which is not expected.")

        # We set the instance properties
        self._b_debug = b_debug
        self._dict_context = dict_context
        self._dict_config = dict_config
        self._session = Session(dash_application=self)

        _the_asset_folder = os.path.join(os.path.dirname(__file__), self.const_str_relative_path_to_assets)
        if os.path.isdir(_the_asset_folder) is False:
            self._logger.error(f"The asset directory could not be found at '{_the_asset_folder}'. Assets will be "
                               f"ignored.")
            _the_asset_folder = None

        # We instantiate the Dash application
        self._dash_server = Dash(
            name=__name__,
            assets_folder=os.path.join(os.path.dirname(__file__), self.const_str_relative_path_to_assets),
            external_stylesheets=[
                dbc.themes.JOURNAL,
            ],
            title=self.const_str__config_label__str_brand,

            # The below setting is to avoid the following errors:
            # Attempting to connect a callback Input item to component: "widget id"
            # but no components with that id exist in the layout.
            #
            # If you are assigning callbacks to components that are
            # generated by other callbacks (and therefore not in the
            # initial layout), you can suppress this exception by setting
            # `suppress_callback_exceptions=True`.
            suppress_callback_exceptions=True
        )

        # We instantiate the widgets
        self._dict_sheets = self._instantiate_sheets(
            dict_widgets_implementations_repository=self.dict_widgets_implementations_repository,
            dict_config=self.dict_config[self.const_str__config_label__dict_sheets],
            dict_context=self.dict_context,
            session=self.session
        )

        # We eventually instantiate the overall layout to load
        self._dash_server.layout = LayoutAsDiv(
            dict_sheets=self._dict_sheets,
            dict_config=self.dict_config.get(self.const_str__config_label__dict_properties, {}),
            session=self.session,
            dict_context=dict_context
        )

        self._dash_server.run_server(debug=True, use_reloader=False)
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def _instantiate_sheets(
            self,
            dict_widgets_implementations_repository: dict[str, Type[Widget]],
            dict_config: dict,
            dict_context: dict,
            session: Session
    ) -> dict[str, SheetAsDiv]:
        """
        Function that instantiates the sheets:
        - The configuration is the overall configuration of the overall dash application
        - The context is the overall Click context, which is not reduced to the parameters expected by the widget.

        The logic is the following:
        - For each widget from "dict_config['widgets']"
        - Get the widget id, which is the key
        - From this widget id, get the associated class from "dict_repository_of_widgets"
        - Instantiate the widget, and add it to the dictionary to return

        :param dict_widgets_implementations_repository: the repository of the widgets, consisting in a widget id
          (similar to the id in the widget json schema
        :param dict_config: the dash application configuration, that contains a node "widgets" in which all the widgets
          to be loaded are listed, along with their configuration
        :param dict_context: the click context that contains all the loaded objects
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        dict_the_return: dict[str, SheetAsDiv] = {}

        # We loop through each configured sheet
        for k_sheet, v_sheet in dict_config.items():
            dict_the_return[k_sheet] = self._instantiate_sheet(
                dict_widgets_implementations_repository=dict_widgets_implementations_repository,
                dict_config=v_sheet,
                dict_context=dict_context,
                str_sheet_name=k_sheet,
                session=session
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return dict_the_return

    def _instantiate_sheet(
            self,
            dict_widgets_implementations_repository: dict[str, Type[Widget]],
            dict_config: dict,
            dict_context: dict,
            str_sheet_name: str,
            session: Session
    ) -> SheetAsDiv:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        lst_the_widgets: list[Widget] = self._instantiate_widgets(
            dict_widgets_implementations_repository=dict_widgets_implementations_repository,
            list_widgets_configs=dict_config.get(self.const_str__config_label__list_widgets, []),
            dict_context=dict_context,
            session=session
        )

        sheet_the_return: SheetAsDiv = SheetAsDiv(
            lst_widgets=lst_the_widgets,
            str_label=str_sheet_name,
            dict_properties=dict_config.get(self.const_str__config_label__dict_properties, {})
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return sheet_the_return

    def _instantiate_widgets(
            self,
            dict_widgets_implementations_repository: dict[str, Type[Widget]],
            list_widgets_configs: list[dict],
            dict_context: dict,
            session: Session
    ) -> list[Widget]:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        lst_the_return: list[Widget] = []

        for i_widget_config in list_widgets_configs:
            lst_the_return.append(
                self._instantiate_widget(
                    dict_widgets_implementations_repository=dict_widgets_implementations_repository,
                    dict_config=i_widget_config,
                    dict_context=dict_context,
                    session=session
                )
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return lst_the_return

    def _instantiate_widget(
            self,
            dict_widgets_implementations_repository: dict[str, Type[Widget]],
            dict_config: dict,
            dict_context: dict,
            session: Session
    ) -> Widget:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if dict_config[self.const_str__config_label__str_widget_id] not in\
                dict_widgets_implementations_repository:
            raise Exception(
                f"Widget id '{dict_config[self.const_str__config_label__str_widget_id]}' does not exist... Only "
                f"the following widgets exist: '{', '.join(dict_widgets_implementations_repository)}'.")

        widget_the_return: Widget = \
            dict_widgets_implementations_repository[dict_config[self.const_str__config_label__str_widget_id]](
                dict_context=dict_context,
                dict_config=dict_config.get(self.const_str__config_label__dict_properties, {}),
                session=session
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return widget_the_return



from inspect import stack
from logging import Logger, getLogger
from pathlib import Path

from pandas import ExcelWriter, DataFrame, read_excel

from com.enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class DatePredictabilityPersister:
    _logger: Logger = getLogger(__name__)

    # sheets' names in the excel spreadsheet
    str_sheet_name__by_measure: str = "by_measure"
    str_sheet_name__by_key: str = "by_key"
    str_sheet_name__historical: str = "historical"

    def persist(
            self,
            obj_bean_to_persist: PredictabilityBean,
            p_file_path: Path,
            str_sheet_name__by_measure: str = str_sheet_name__by_measure,
            str_sheet_name__by_key: str = str_sheet_name__by_key,
            str_sheet_name__historical: str = str_sheet_name__historical,
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We check there is no existing file at the provided path
        # Note: if simlink, exists returns whether or not the simlink points to an existing file or directory
        if p_file_path.exists() | p_file_path.is_symlink():
            raise Exception(
                f"Cannot persist instance of PredictabilityBean, as provided path '{str(p_file_path)}' seems to exist:"
                f"\n\t- is_file: {p_file_path.is_file()}"
                f"\n\t- is_dir: {p_file_path.is_dir()}"
                f"\n\t- is_symlink: {p_file_path.is_symlink()}"
            )

        # Creating Excel Writer Object from Pandas
        with ExcelWriter(str(p_file_path)) as obj_the_writer:

            obj_bean_to_persist.df_by_measure.to_excel(
                obj_the_writer,
                sheet_name=str_sheet_name__by_measure,
                index=False
            )

            obj_bean_to_persist.df_by_key.to_excel(
                obj_the_writer,
                sheet_name=str_sheet_name__by_key,
                index=False
            )

            obj_bean_to_persist.df_historical.to_excel(
                obj_the_writer,
                sheet_name=str_sheet_name__historical,
                index=False
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def load(
            self,
            p_file_path: Path,
            str_sheet_name__by_measure: str = str_sheet_name__by_measure,
            str_sheet_name__by_key: str = str_sheet_name__by_key,
            str_sheet_name__historical: str = str_sheet_name__historical,
    ) -> PredictabilityBean:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We check there is an existing file at the provided path
        if not p_file_path.is_file():
            raise Exception(
                f"Cannot load an instance of PredictabilityBean, as provided path '{str(p_file_path)}' does not seem "
                f"to exist or to be a file:"
                f"\n\t- is_file: {p_file_path.is_file()}"
                f"\n\t- is_dir: {p_file_path.is_dir()}"
                f"\n\t- is_symlink: {p_file_path.is_symlink()}"
            )

        # We load the 3 sheets

        # #################################################
        # 1.1. We load the measures
        try:
            df_the_by_measure: DataFrame = read_excel(
                io=p_file_path,
                sheet_name=str_sheet_name__by_measure
            )
            self._logger.info(f"Sheet {str_sheet_name__by_measure} read, shape '{df_the_by_measure.shape}'.")
        except Exception as obj_the_exception:
            raise Exception(
                f"Could not load sheet '{str_sheet_name__by_measure}' from file '{str(p_file_path)}'. Make sure the "
                f"file is effectively an excel spreadsheet that can be opened."
            ) from obj_the_exception
        # 1.2. We properly type the dataframe
        try:
            df_the_by_measure = PandasDataframeTyper.type(
                df_to_type=df_the_by_measure,
                dict_columns_to_type={
                    DatePredictabilityComputer.str__input__column_label__date: PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__input__column_label__measure: PandasDataframeTyper.str__type__date,
                },
                b_strict=False
            )
        except Exception as an_exception:
            raise Exception(
                f"Could not type sheet '{str_sheet_name__by_measure}' from file '{str(p_file_path)}'. Make sure the "
                f"data is correctly formatted in the file.") from an_exception

        # #################################################
        # 2.1. We load the statistics by key
        try:
            df_the_by_key: DataFrame = read_excel(
                io=p_file_path,
                sheet_name=str_sheet_name__by_key
            )
            self._logger.info(f"Sheet {str_sheet_name__by_key} read, shape '{df_the_by_key.shape}'.")
        except Exception as obj_the_exception:
            raise Exception(
                f"Could not load sheet '{str_sheet_name__by_key}' from file '{str(p_file_path)}'. Make sure the "
                f"file is effectively an excel spreadsheet that can be opened."
            ) from obj_the_exception
        # 2.2. We properly type the dataframe
        try:
            df_the_by_key = PandasDataframeTyper.type(
                df_to_type=df_the_by_key,
                dict_columns_to_type={
                    DatePredictabilityComputer.str__statistics__column_label__measure_count:
                        PandasDataframeTyper.str__type__int,
                    DatePredictabilityComputer.str__statistics__column_label__date_first:
                        PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__statistics__column_label__date_last:
                        PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__statistics__column_label__measure_min:
                        PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__statistics__column_label__measure_max:
                        PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__statistics__column_label__measure_first:
                        PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__statistics__column_label__measure_last:
                        PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__statistics__column_label__predictability_last:
                        PandasDataframeTyper.str__type__float,
                },
                b_strict=False
            )
        except Exception as an_exception:
            raise Exception(
                f"Could not type sheet '{str_sheet_name__by_key}' from file '{str(p_file_path)}'. Make sure the "
                f"data is correctly formatted in the file.") from an_exception

        # #################################################
        # 3.1. We load the historical
        try:
            df_the_historical: DataFrame = read_excel(
                io=p_file_path,
                sheet_name=str_sheet_name__historical
            )
            self._logger.info(f"Sheet {str_sheet_name__historical} read, shape '{df_the_historical.shape}'.")
        except Exception as obj_the_exception:
            raise Exception(
                f"Could not load sheet '{str_sheet_name__historical}' from file '{str(p_file_path)}'. Make sure the "
                f"file is effectively an excel spreadsheet that can be opened."
            ) from obj_the_exception
        # 3.2. We properly type the dataframe
        try:
            df_the_historical = PandasDataframeTyper.type(
                df_to_type=df_the_historical,
                dict_columns_to_type={
                    DatePredictabilityComputer.str__input__column_label__date: PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__input__column_label__measure: PandasDataframeTyper.str__type__date,
                    DatePredictabilityComputer.str__output__column_label__predictability:
                        PandasDataframeTyper.str__type__float,
                },
                b_strict=False
            )
        except Exception as an_exception:
            raise Exception(
                f"Could not type sheet '{str_sheet_name__historical}' from file '{str(p_file_path)}'. Make sure the "
                f"data is correctly formatted in the file.") from an_exception

        # We instantiate the bean to return
        obj_the_return: PredictabilityBean = PredictabilityBean(
            df_by_measure=df_the_by_measure,
            df_by_key=df_the_by_key,
            df_historical=df_the_historical
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        return obj_the_return



import logging
from inspect import stack

import click
from click import BadParameter
from pandas import DataFrame

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp-compute')
@click.pass_context
@click.argument('alias-in', type=str)
@click.argument('alias-out', type=str)
@click.option(
    '--resample/--no-resample', type=bool, default=False,
    help='To resample daily the measures, to have better graphing.',
)
@click.option('-k', '--key', help="The key column, that will be renamed.", type=str, default=None)
@click.option('-d', '--date', help="The date column, that will be renamed.", type=str, default=None)
@click.option('-m', '--measure', help="The measure column, that will be renamed.", type=str, default=None)
def dp_compute_date_predictability(ctx_context, alias_in, alias_out, resample, key, date, measure):
    """
    Compute date predictability for a dataframe that is to contain 3 columns, and 3 columns only:
    - One column named DatePredictabilityComputer.str__input__column_label__key, aka "key"
    - One column named DatePredictabilityComputer.str__input__column_label__date, aka "date"
    - One column named DatePredictabilityComputer.str__input__column_label__measure, aka "measure".

    The following business rules are checked:
    - BR_001, ALIAS_IN should exist in context
    - BR_002, ALIAS_IN should be a dataframe
    - BR_003, ALIAS_OUT exists in context --> Warning

    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    # BR_001, ALIAS_IN should exist in context
    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"BR_001, no object labelled '{alias_in}' could be found in context, which is not expected.",
            param_hint=f"BR_001, ensure you effectively have an object labelled '{alias_in}' in context.",
            ctx=ctx_context
        )

    # BR_002, ALIAS_IN should be a dataframe
    if not isinstance(ctx_context.obj[alias_in], DataFrame):
        raise BadParameter(
            message=f"BR_002, object labelled '{alias_in}' is of type '{type(ctx_context.obj[alias_in])}', which is not"
                    f" expected.",
            param_hint=f"Ensure to provide an 'DataFrame' instance.",
            ctx=ctx_context
        )

    # BR_003, ALIAS_OUT exists in context --> Warning
    if alias_out in ctx_context.obj:
        _logger.warning(
            f"BR_006, warning: another object with alias '{alias_out}' already exists, of type "
            f"'{type(ctx_context.obj[alias_in])}'. This data will be overridden."
        )

    df_the_input: DataFrame = ctx_context.obj[alias_in]

    if key is not None:
        df_the_input = df_the_input.rename(columns={key: DatePredictabilityComputer.str__input__column_label__key})
    if date is not None:
        df_the_input = df_the_input.rename(columns={date: DatePredictabilityComputer.str__input__column_label__date})
    if measure is not None:
        df_the_input = df_the_input.rename(
            columns={measure: DatePredictabilityComputer.str__input__column_label__measure})

    obj_the_computer: DatePredictabilityComputer = DatePredictabilityComputer()

    obj_the_bean: PredictabilityBean = obj_the_computer.compute_historical_date_predictability(
        df_measures=df_the_input,
        b_resample=resample
    )

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = obj_the_bean
    else:
        ctx_context.obj[alias_in] = obj_the_bean

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger
from datetime import timedelta, datetime

from pandas import Series, DataFrame

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class DataToGraph:
    _str_figure_title: str
    _str_abscissa_label: str

    _str_predictabilities_ordinate_label: str  # the label for the ordinate axis
    _s_predictabilities_values: Series  # the ordinate predictability values to print in the graph
    _s_predictabilities_dates: Series  # the abscissa dates to print in the graph
    _str_predictabilities_name: str  # the trace name, displayed in the legend

    _str_measures_ordinate_label: str  # the label for the ordinate axis
    _s_measures_values: Series  # the ordinate measure values to print in the graph
    _s_measures_dates: Series  # the abscissa dates to print in the graph
    _str_measures_name: str  # the trace name, displayed in the legend

    # the cone in the measure widget
    _lst_increasing_cone_values: list[datetime]
    _lst_increasing_cone_dates: list[datetime]
    _str_increasing_cone_name: str
    _lst_decreasing_cone_values: list[datetime]
    _lst_decreasing_cone_dates: list[datetime]
    _str_decreasing_cone_name: str
    _lst_last_measure_values: list[datetime]
    _lst_last_measure_dates: list[datetime]
    _str_last_measure_name: str

    @property
    def str_figure_title(self) -> str:
        return self._str_figure_title

    @property
    def str_abscissa_label(self) -> str:
        return self._str_abscissa_label

    @property
    def str_predictabilities_ordinate_label(self) -> str:
        return self._str_predictabilities_ordinate_label

    @property
    def s_predictabilities_values(self) -> Series:
        return self._s_predictabilities_values

    @property
    def s_predictabilities_dates(self) -> Series:
        return self._s_predictabilities_dates

    @property
    def str_predictabilities_name(self) -> str:
        return self._str_predictabilities_name

    @property
    def str_measures_ordinate_label(self) -> str:
        return self._str_measures_ordinate_label

    @property
    def s_measures_values(self) -> Series:
        return self._s_measures_values

    @property
    def s_measures_dates(self) -> Series:
        return self._s_measures_dates

    @property
    def str_measures_name(self) -> str:
        return self._str_measures_name

    @property
    def lst_increasing_cone_values(self) -> list[datetime]:
        return self._lst_increasing_cone_values

    @property
    def lst_increasing_cone_dates(self) -> list[datetime]:
        return self._lst_increasing_cone_dates

    @property
    def str_increasing_cone_name(self) -> str:
        return self._str_increasing_cone_name

    @property
    def lst_decreasing_cone_values(self) -> list[datetime]:
        return self._lst_decreasing_cone_values

    @property
    def lst_decreasing_cone_dates(self) -> list[datetime]:
        return self._lst_decreasing_cone_dates

    @property
    def str_decreasing_cone_name(self) -> str:
        return self._str_decreasing_cone_name

    @property
    def lst_last_measure_values(self) -> list[datetime]:
        return self._lst_last_measure_values

    @property
    def lst_last_measure_dates(self) -> list[datetime]:
        return self._lst_last_measure_dates

    @property
    def str_last_measure_name(self) -> str:
        return self._str_last_measure_name

    _logger: Logger = getLogger(__name__)

    def _compute_cones_coordinates(
            self,
            df_the_historical_data: DataFrame
    ) -> tuple[
        list[datetime],  # _lst_increasing_cone_values
        list[datetime],  # _lst_increasing_cone_dates
        list[datetime],  # _lst_decreasing_cone_values
        list[datetime],  # _lst_decreasing_cone_dates
        list[datetime],  # _lst_last_measure_values
        list[datetime]   # _lst_last_measure_dates
    ]:
        """
        From a given historical dataframe, the function returns the following variables:
        - _lst_increasing_cone_values
        - _lst_increasing_cone_dates
        - _lst_decreasing_cone_values
        - _lst_decreasing_cone_dates
        - _lst_last_measure_values
        - _lst_last_measure_dates
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_stats: DataFrame = DatePredictabilityComputer().compute_statistics_by_key(
            df_measures=df_the_historical_data,
            df_historical_predictability=df_the_historical_data,
        )

        if len(df_the_stats.index) != 1:
            raise Exception(f"The function DatePredictabilityComputer.compute_statistics_by_key returned "
                            f"'{len(df_the_stats.index)}' rows, while '1' one expected.")

        # We retrieve the stats as a directory
        d_the_stats: dict = df_the_stats.iloc[0, :].to_dict()

        dt_the_bottom_left_corner: datetime = min(
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last],
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__date_first]
        )
        dt_the_top_right_corner: datetime = max(
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last],
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__date_last]
        )
        dt_the_top_left_corner: datetime = max(
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last],
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last]
            + timedelta(
                days=(
                        d_the_stats[
                            DatePredictabilityComputer.str__statistics__column_label__measure_last]
                        - d_the_stats[
                            DatePredictabilityComputer.str__statistics__column_label__date_first]
                ).days
            )
        )
        dt_the_bottom_right_corner: datetime = min(
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last],
            d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last]
            - timedelta(
                days=(
                        d_the_stats[
                            DatePredictabilityComputer.str__statistics__column_label__date_last]
                        - d_the_stats[
                            DatePredictabilityComputer.str__statistics__column_label__measure_last]
                ).days
            )
        )

        lst_the_increasing_cone_values: list[datetime] = [dt_the_bottom_left_corner, dt_the_top_right_corner]
        lst_the_increasing_cone_dates: list[datetime] = [dt_the_bottom_left_corner, dt_the_top_right_corner]

        lst_the_decreasing_cone_values: list[datetime] = [dt_the_top_left_corner, dt_the_bottom_right_corner]
        lst_the_decreasing_cone_dates: list[datetime] = [dt_the_bottom_left_corner, dt_the_top_right_corner]

        lst_the_last_measure_values: list[datetime] = \
            [d_the_stats[DatePredictabilityComputer.str__statistics__column_label__measure_last]]*2
        lst_the_last_measure_dates: list[datetime] = [dt_the_bottom_left_corner, dt_the_top_right_corner]

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return \
            lst_the_increasing_cone_values, lst_the_increasing_cone_dates, \
            lst_the_decreasing_cone_values, lst_the_decreasing_cone_dates, \
            lst_the_last_measure_values, lst_the_last_measure_dates

    def __init__(
            self,
            obj_predictability: PredictabilityBean,
            str_filter_on_key: str,
            dt_as_of_date: datetime = None
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We set the figures titles and labels
        self._str_figure_title = f"Date Predictability for <b>{str_filter_on_key}</b>"
        self._str_abscissa_label = DatePredictabilityComputer.str__input__column_label__date
        self._str_predictabilities_ordinate_label = DatePredictabilityComputer.str__output__column_label__predictability
        self._str_measures_ordinate_label = DatePredictabilityComputer.str__input__column_label__measure

        # We select the data for that selected key
        df_the_historical_data: DataFrame = obj_predictability.df_historical[
            obj_predictability.df_historical[
                DatePredictabilityComputer.str__input__column_label__key
            ] == str_filter_on_key
        ]

        # If we do not zoom on a given date, we set it to the maximum (as later filter will always consider this date)
        if dt_as_of_date is None:
            dt_as_of_date = df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date].max()

        # We prepare the data for the historical evolution of the predictability
        self._s_predictabilities_values = \
            df_the_historical_data[DatePredictabilityComputer.str__output__column_label__predictability]
        self._s_predictabilities_dates = \
            df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date]
        self._str_predictabilities_name = DatePredictabilityComputer.str__output__column_label__predictability

        # We prepare the data for the historical evolution of the measures
        self._s_measures_values = \
            df_the_historical_data[
                df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date] <= dt_as_of_date
            ][DatePredictabilityComputer.str__input__column_label__measure]
        self._s_measures_dates = \
            df_the_historical_data[
                df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date] <= dt_as_of_date
            ][DatePredictabilityComputer.str__input__column_label__date]
        self._str_measures_name = DatePredictabilityComputer.str__input__column_label__measure

        # We prepare the data for displaying the convergence cone alongside the measures
        self._lst_increasing_cone_values, \
            self._lst_increasing_cone_dates, \
            self._lst_decreasing_cone_values, \
            self._lst_decreasing_cone_dates, \
            self._lst_last_measure_values, \
            self._lst_last_measure_dates = self._compute_cones_coordinates(
                df_the_historical_data=df_the_historical_data[
                    df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date] <= dt_as_of_date
                ]
            )
        self._str_increasing_cone_name = "Increasing convergence cone"
        self._str_decreasing_cone_name = "Decreasing convergence cone"
        self._str_last_measure_name = "Final measure"

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")


class DataToTable:
    _dict_values: dict[str, ]

    @property
    def dict_values(self) -> dict[str, ]:
        return self._dict_values

    _logger: Logger = getLogger(__name__)

    def __init__(
            self,
            obj_predictability: PredictabilityBean,
            str_filter_on_key: str,
            dt_as_of_date: datetime = None
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # If we do not zoom on a given date, we just get the statistics already available
        if dt_as_of_date is None:
            # We select the data for that selected key
            df_the_statistics_data: DataFrame = obj_predictability.df_by_key[
                obj_predictability.df_by_key[
                    DatePredictabilityComputer.str__input__column_label__key
                ] == str_filter_on_key
            ][DatePredictabilityComputer.lst__output__statistics_by_key__column_labels]  # Required to order columns

        # Otherwise, we need to compute the statistics
        else:

            # We select the data for that selected key, and date
            df_by_measure_data: DataFrame = obj_predictability.df_historical[
                (
                    obj_predictability.df_historical[
                        DatePredictabilityComputer.str__input__column_label__key
                    ] == str_filter_on_key
                )
                & (
                    obj_predictability.df_historical[
                        DatePredictabilityComputer.str__input__column_label__date
                    ] <= dt_as_of_date
                )
            ]
            df_historical_data: DataFrame = obj_predictability.df_historical[
                (
                    obj_predictability.df_historical[
                        DatePredictabilityComputer.str__input__column_label__key
                    ] == str_filter_on_key
                )
                & (
                    obj_predictability.df_historical[
                        DatePredictabilityComputer.str__input__column_label__date
                    ] <= dt_as_of_date
                )
            ]

            df_the_statistics_data: DataFrame = DatePredictabilityComputer().compute_statistics_by_key(
                    df_measures=df_by_measure_data,
                    df_historical_predictability=df_historical_data
            )

        # We expect only one record
        if len(df_the_statistics_data.index) != 1:
            raise Exception(f"When filtering statistics on key '{str_filter_on_key}', we got "
                            f"'{len(df_the_statistics_data.index)}' line while one and only one was expected.")

        # self._lst_values = df_the_statistics_data.iloc[0, :].tolist()
        self._dict_values = df_the_statistics_data.to_dict("records")[0]

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                           f"'{self._dict_values}'.")



from os.path import join, dirname
from inspect import stack
from logging import Logger, getLogger
import json
from pathlib import Path
from typing import Union

from jsonschema import RefResolver, Validator
from jsonschema.validators import validator_for


class ConfigChecker:
    """
    The ConfigChecker
    """
    _logger: Logger = getLogger(__name__)

    # The main excel_dashboard json schema
    _path_excel_dashboard_json_schema: Path = Path(join(dirname(__file__), 'config_checkers/excel_dashboard.json'))

    # Basic json schemas, used across widgets, are subscribed below.
    _lst_paths_to_basic_json_schemas: list = [
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_workbook__add_format.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_colors.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_vba.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_address_range.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_worksheet__add_table.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_worksheet__conditional_format__options.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_worksheet__conditional_format.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_workbook__set_properties.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_worksheet__freeze_panes.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_worksheet__set_column.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_worksheet__set_row.json')),
        Path(join(dirname(__file__), 'config_checkers/xlsxwriter_workbook__options.json')),
    ]

    # The Validator instance which is used to later validate json configurations
    _obj_validator: Validator = None

    def __init__(
            self,
            lst_widgets_json_schemas: list[dict],
            str_base_json_schema_id: str = "https://enovation.com/excel_dashboard",
    ):
        """
        Initialize the config checker. The json schemas are loaded, and a RefResolver is instantiated.

        By default, the config checker will validate an excel dashboard, but we can set another json schema, mostly
        used to validate json schemas unitarily.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _dict_the_store_of_json_schema: dict = {}

        # We register the widgets' schemas: their schemas are loaded, and registered in the dictionary
        for i_json_schema_path in lst_widgets_json_schemas:
            self._register_json_schema(
                dict_store_of_json_schema=_dict_the_store_of_json_schema,
                json_schema=i_json_schema_path
            )

        # We register the the main excel dashboard schema, after having enriched it with the eligible widgets
        _dict_the_enriched_excel_dashboard_json_schema: dict = self._enrich_master_json_schema(
            dict_store_of_json_schema=_dict_the_store_of_json_schema,
            path_json_schema=self._path_excel_dashboard_json_schema
        )

        self._register_json_schema(
            dict_store_of_json_schema=_dict_the_store_of_json_schema,
            json_schema=_dict_the_enriched_excel_dashboard_json_schema
        )

        # We register the basic json schemas
        for i_json_schema_path in self._lst_paths_to_basic_json_schemas:
            self._register_json_schema(
                dict_store_of_json_schema=_dict_the_store_of_json_schema,
                json_schema=i_json_schema_path
            )

        # We eventually instantiate the reference resolver, and the Validator
        if str_base_json_schema_id not in _dict_the_store_of_json_schema:
            raise Exception(f"We could not find the base json schema '{str_base_json_schema_id}'.")

        _the_resolver: RefResolver = RefResolver.from_schema(
            schema=_dict_the_store_of_json_schema[str_base_json_schema_id],
            store=_dict_the_store_of_json_schema
        )

        self._obj_validator = validator_for(
            _dict_the_store_of_json_schema[str_base_json_schema_id]
        )(
            _dict_the_store_of_json_schema[str_base_json_schema_id],
            resolver=_the_resolver
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def _register_json_schema(
            self,
            dict_store_of_json_schema: dict,
            json_schema: Union[Path, dict],
    ):
        """
        Function that:
         - Loads a json schema from a path
         - checks there is an $id
         - checks this $id was not previously registered, and
         - eventually load this schema into the dictionary.
        :param dict_store_of_json_schema: the store of json schemas
        :param json_schema: the json schema which is processed, either a path to a json file or a dictionary
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _dict_json_schema: dict
        _error_msg: str

        if isinstance(json_schema, Path):
            with open(json_schema, "r") as _the_file:
                _dict_json_schema: dict = json.loads(_the_file.read())
            _error_msg = f"When loading json schema from file '{json_schema}'"
        else:
            _dict_json_schema = json_schema
            _error_msg = f"When loading json schema '{json_schema}'"

        if "$id" not in _dict_json_schema:
            self._logger.error(f"{_error_msg}', it did not have an $id, so it could not be loaded in the json schemas "
                               f"store.")
        elif _dict_json_schema["$id"] in dict_store_of_json_schema:
            self._logger.error(f"{_error_msg}, its $id '{_dict_json_schema['$id']}' was already registered. This new "
                               f"version will be therefore ignored, and only the first version will be kept.")
        else:
            dict_store_of_json_schema[_dict_json_schema["$id"]] = _dict_json_schema

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def _enrich_master_json_schema(
            self,
            dict_store_of_json_schema: dict,
            path_json_schema: Path
    ) -> dict:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        with open(path_json_schema, "r") as _the_file:
            _dict_the_return: dict = json.loads(_the_file.read())

        _lst_widgets_schemas: list[dict] = [
            {
                "properties": {
                    "widget_id": {
                        "const": i_schema_id
                    },
                    "config": {
                        "$ref": i_schema_id[len("https://enovation.com"):]
                    }
                },
                "required": [
                    "config"
                ]
            }
            for i_schema_id in dict_store_of_json_schema
        ]

        _dict_the_return["properties"]["sheets"]["patternProperties"][""]["properties"]["widgets"][
            "patternProperties"][""]["oneOf"] = _lst_widgets_schemas

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return _dict_the_return

    def validate(self, dict_to_validate: dict):
        """
        Validate the parameter "dict_to_validate" against a given json schema. If the dict to validate is correct, the
        function returns. Otherwise, an exception ValidationError is thrown.

        In case we need to make the ValidationError more explicit, we can override this function, catch the exception,
        and adjust it to make it more meaningful before re-raising it.

        :param dict_to_validate: the dictionary to validate
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        self._obj_validator.validate(
            instance=dict_to_validate
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack
from logging import Logger, getLogger

from dash import html, dcc
import dash_bootstrap_components as dbc
from dash.html import Div

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_session import Session
import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_wid as wid


class TabExploreAsDiv(Div):
    """
    """
    _logger: Logger = getLogger(__name__)

    def __init__(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # ##############################################################################################################
        # Callbacks should be imported here... Otherwise, we face "unknown id" as callbacks get registered even though
        # the widgets are not used...
        # noinspection PyUnresolvedReferences
        import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_p2_explore.\
            tab_explore_callbacks
        # ##############################################################################################################

        # We retrieve the predictability bean
        obj_the_predictability_bean: PredictabilityBean = Session.get_predictability_bean()

        super().__init__(
            children=[
                # A title
                html.H3("Zoom on a given key"),

                # The filter to select a key (aka. a project, an opportunity, etc) to graph and zoom in
                dcc.Dropdown(
                    id=wid.str_the_wid__explore__key__as_drop_down,
                    placeholder="Select a key to graph and zoom in...",
                    options=[
                        {'label': x, 'value': x}
                        for x in
                        obj_the_predictability_bean.df_by_key[DatePredictabilityComputer.str__input__column_label__key]
                    ],
                ),

                html.Div(
                    children=[
                        html.Div(
                            id=wid.str_the_wid__explore__graph__as_div,
                            # by default, no key is selected, so the graph and table are not displayed
                            # style={"display": "none"},
                            children=[
                                dcc.Graph(id=wid.str_the_wid__explore__graph__as_graph)
                            ]
                        ),
                        html.Div(
                            id=wid.str_the_wid__explore__table__as_div,
                            children=[
                                # The table
                                dbc.Table(
                                    html.Tbody([
                                        html.Tr(
                                            [
                                                html.Td(i_col),
                                                html.Td(
                                                    "--",
                                                    id=(
                                                            wid.str_the_wid__explore__table__prefix_to_access_columns +
                                                            i_col +
                                                            wid.str_the_wid__explore__table__suffix_to_access_columns
                                                    )
                                                )
                                            ]
                                        )
                                        for i_col in obj_the_predictability_bean.df_by_key.columns
                                    ]),
                                    color="primary",
                                    bordered=True
                                )
                            ],
                            # by default, no key is selected, so the graph and table are not displayed
                            # style={"display": "none"}
                        ),
                    ]
                )
            ],
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from dash import html, dcc
from inspect import stack
from logging import Logger, getLogger

from dash import callback, Output, Input
from dash.html import P

from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.widget import WidgetAsDiv


class DebugAsDiv(WidgetAsDiv):
    _logger: Logger = getLogger(__name__)

    # The widget IDs to be used
    # Naming convention:
    # - should start by "const_str__wid__"
    # - as the application will screen all the widget ids, across widgets, to ensure there is no ambiguity, aka same
    #   widget id across different widgets
    const_str__wid__debug_as_div: str = "enov-content__debug__as_div"
    const_str__wid__debug_as_samp: str = "enov-content__debug__as_samp"
    const_str__wid__debug_as_interval: str = "enov-content__debug__as_interval"
    const_str__wid__debug_as_dcc_store: str = "enov-content__debug__as_dcc_store"

    """
    Return the welcome widget such as:
    <div class="jumbotron">
        <h1>Date Predictability</h1>
        <p class="lead">
            'Predictability' is about accurately predicting early enough the project outcomes (incl. beyond own
            roles and responsibilities), to enable timely corrective actions that increase the likelihood of
            achieving targets and reducing outcome variance.
        </p>
        <hr class="my-2">
        <p>Explanation of how it works here...</p>
    </div>
    """

    def __init__(
            self,
            dict_config: dict,
            dict_context: dict,
            session: Session,
            *args, **kwargs
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        super().__init__(
            dict_config=dict_config,
            dict_context=dict_context,
            session=session,
            children=[
                html.Div(
                    children="Debug",
                    className="bg-secondary",
                    style={
                        "margin-left": "-0.75rem",
                        "margin-right": "-0.75rem",
                        "padding-left": "0.75rem",
                    }
                ),
                html.Div(
                    children=[
                        html.Samp(
                            children=[],
                            style={"font-size": "50%"},
                            id=self.const_str__wid__debug_as_samp,
                        ),
                    ]
                ),
                dcc.Interval(
                    id=self.const_str__wid__debug_as_interval,
                    interval=1000,  # in milliseconds
                    n_intervals=0
                )
            ],
            id=self.const_str__wid__debug_as_div,
            className="container border bg-light",
            style={"margin-top": "1rem"}
        )

        @callback(
            Output(self.const_str__wid__debug_as_samp, 'children'),
            Input(self.const_str__wid__debug_as_interval, 'n_intervals'),
        )
        def _callback_debug(i_interval_n: int) -> list[P]:
            """
            Callback function to update the debug
            :param i_interval_n: number of seconds since the application started
            :return:
            """
            self._logger.debug(
                f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameter i_interval_n = "
                f"{i_interval_n}.")

            _lst_str_debug: list[str] = self.session.lst_str_debug_trace

            _lst_the_return: list[P] = [
                P(
                    f">> {i_str}",
                    className="m-0 m-0 p-0"
                ) for i_str in _lst_str_debug
            ]
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

            return _lst_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack

import click
from click import BadParameter
from pandas import DataFrame
import logging

from com.enovation.helper.click.python_literal_argument_and_option import PythonLiteralArgument, PythonLiteralOption
from com.enovation.helper.pandas_dataframe_sampler import PandasDataframeSampler

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('df-rename-columns')
@click.pass_context
@click.argument('alias-in', type=str)
@click.argument('dictionary', type=list, cls=PythonLiteralArgument)
@click.option('-o', '--alias-out', help="Resulting dataframe.", type=str)
def df_rename_columns(ctx_context, alias_in, dictionary, alias_out):
    """
    Rename dataframe columns.

    :param ctx_context: Click context.
    :param alias_in: dataframe to lad from the context, so its columns can be renamed.
    :param dictionary: dictionary to rename columns, in the form '{"col1":"new col1", ... , "col-x":"new col-x"}'.
    :param alias_out: alias to the cleansed dataframe, in case source dataframe should not be altered.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot rename columns from '{alias_in}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively have a dataframe labelled '{alias_in}' loaded in context.",
            ctx=ctx_context
        )
    if alias_out in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias_out}' already exists, of type {type(alias_out)}. This data "
            f"will be overridden."
        )

    df_the_input: DataFrame = ctx_context.obj[alias_in]

    if not isinstance(df_the_input, DataFrame):
        raise BadParameter(
            message=f"Cannot rename columns from '{alias_in}', as this alias is of type '{type(df_the_input)}' while "
                    f"'DataFrame' is expected.",
            param_hint=f"Ensure you effectively assigned a DataFrame to the alias '{alias_in}'.",
            ctx=ctx_context
        )

    # JSG, as of 28-Oct-2021: introduced PythonLiteralArgument
    # So the below is not required anymore...
    # try:
    #     dict_the_dictionary: dict = json.loads(dictionary)
    # except json.decoder.JSONDecodeError as error:
    #     raise BadParameter(
    #         message=f"Cannot transform '{dictionary}' as a dictionary to rename columns.",
    #         param_hint=f"Ensure you effectively provided a string with 'key':'value' assignment that can be used to "
    #                    f"rename columns.",
    #         ctx=ctx_context
    #     ) from error

    # We check the dictionary is a simple set of key:value assignments
    lst_the_errors: list = [
        "'"+k+"'" for k, v in dictionary.items()
        if (
            (not isinstance(k, str))
            | (not isinstance(v, str))
        )
    ]
    if len(lst_the_errors):
        raise BadParameter(
            message=f"Cannot use '{dictionary}' as a dictionary to rename columns.",
            param_hint=f"Ensure you effectively provided a string with simple 'key as string':'value as string' "
                       f"assignment that can be used to rename columns.",
            ctx=ctx_context
        )

    # We check the dictionary only provides keys that effectively exist as columns
    lst_the_errors: list = list({k for k in dictionary}-set(df_the_input.columns))
    if len(lst_the_errors) > 0:
        raise BadParameter(
            message=f"Columns '{', '.join(lst_the_errors)}' are to be renamed, but they do not exist in the dataframe.",
            param_hint=f"Ensure you effectively provide columns to rename that exist in the dataframe.",
            ctx=ctx_context
        )

    df_the_return: DataFrame = df_the_input.rename(columns=dictionary)

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = df_the_return
    else:
        ctx_context.obj[alias_in] = df_the_return

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('df-cleanse-null')
@click.pass_context
@click.argument('alias-in', type=str)
@click.option('-c', '--columns', type=list, cls=PythonLiteralOption, default=None,
              help="The columns to screen for null values.")
@click.option('-v', '--value', help="Value to replace null.", type=str)
@click.option('-o', '--alias-out', help="Resulting dataframe.", type=str)
def df_cleanse_null(ctx_context, alias_in, columns, value, alias_out):
    """
    Cleanses null values from the dataframe:
    - In case VALUE is not set, we drop the rows, or we replace null values with VALUE
    - In case COLUMNS is not set, we process all columns from the dataframe, or we only process the columns listed.
    """
    _logger.info(f"\n"
                 f"\n**************************************************************************************************"
                 f"\n**************************************************************************************************"
                 f"\nCommand mx-sp-pp-load-history called with parameters:"
                 f"\n\t- alias-in: {alias_in}"
                 # f"\n\t- columns: {', '.join(columns)}"
                 f"\n\t- value: {value}"
                 f"\n\t- alias-out: {alias_out}")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot cleanse '{alias_in}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively loaded some data, and labelled these with the alias '{alias_in}'.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias_in], DataFrame):
        raise BadParameter(
            message=f"Cannot cleanse '{alias_in}', as this alias is of type '{type(ctx_context.obj[alias_in])}' while "
                    f"'DataFrame' is expected.",
            param_hint=f"Ensure you effectively used operands to generate data as a DataFrame.",
            ctx=ctx_context
        )

    if alias_out in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias_out}' already exists, of type {type(alias_out)}. This data "
            f"will be overridden."
        )

    df_the_measures = ctx_context.obj[alias_in]

    # If the columns are not listed
    if columns is None:
        columns = list(df_the_measures.columns)

    for i_col in columns:
        if i_col not in df_the_measures.columns:
            raise Exception(f"The command cleanse_null_values cannot access column '{i_col}'. First, ensure this "
                            f"column effectively exists (aka. present among the columns of a loaded file, or present "
                            f"among generated data by a command), then call this function to cleanse null values.")

        if value:
            df_the_measures = df_the_measures[i_col].fillna(value)
        else:
            df_the_measures = df_the_measures[df_the_measures[i_col].notna()]
            _logger.info(f"Rows containing null value in column '{i_col}' deleted, shape '{df_the_measures.shape}'.")

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = df_the_measures
    else:
        ctx_context.obj[alias_in] = df_the_measures

    _logger.info(f"\nCommand df-cleanse-null is returning, after having enriched the context with:"
                 f"\n\t- '{alias_out}' dataframe: {df_the_measures.shape}"
                 f"\n\t- containing columns: {', '.join(df_the_measures.columns)}"
                 f"\n**************************************************************************************************"
                 f"\n**************************************************************************************************"
                 f"\n"
                 )


@click.option('-k', '--key-columns', type=list, cls=PythonLiteralOption, default='[]',
              help="Columns labels to isolate independent universes.")
@click.option('-r', '--ordering-columns', type=list, cls=PythonLiteralOption, default='[]',
              help="Columns labels to order the dataframe, before comparing rows in the right order.")
@click.option('-v', '--value-columns', type=list, cls=PythonLiteralOption, default=None,
              help="Columns labels to compare rows, and identify duplicates to drop.")
@click.option('--keep-last/--no-keep-last', type=bool, default=False,
              help='Whether or not we keep the last record, even if duplicate with previous rows.')
@click.argument('alias-in', type=str)
@click.option('-o', '--alias-out', help="Resulting dataframe.", type=str)
@click.command('df-compress')
@click.pass_context
def df_compress(ctx_context, alias_in, alias_out, key_columns, ordering_columns, value_columns, keep_last):
    """
    The function screens the rows from Dataframe ALIAS-IN, universe by universe as defined from KEY-COLUMNS, with
    rows ordered by ORDERING-COLUMNS, and will drop all rows that have all VALUE-COLUMNS identical.

    If argument b_keep_last is set to True, the very last record of each universe is kept, even if it is a duplicate
    with a previous row.

    Several business rules are checked:
    - BR_001, dataframe expected not to contain several columns with the same label
    - BR_002, all columns from lst_key_columns expected to be in the dataframe
    - BR_003, all columns from lst_value_columns expected to be in the dataframe
    - BR_004, all columns from lst_ordering_columns expected to be in the dataframe
    - BR_005, all columns from dataframe should be listed into one of the list: key, value or ordering
    - BR_006, dataframe expected not to contain any duplicate across the ordering columns (within the same universe)
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot rename columns from '{alias_in}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively have a dataframe labelled '{alias_in}' loaded in context.",
            ctx=ctx_context
        )
    if alias_out in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias_out}' already exists, of type {type(alias_out)}. This data "
            f"will be overridden."
        )

    df_the_input: DataFrame = ctx_context.obj[alias_in]

    if not isinstance(df_the_input, DataFrame):
        raise BadParameter(
            message=f"Cannot rename columns from '{alias_in}', as this alias is of type '{type(df_the_input)}' while "
                    f"'DataFrame' is expected.",
            param_hint=f"Ensure you effectively assigned a DataFrame to the alias '{alias_in}'.",
            ctx=ctx_context
        )

    df_the_return: DataFrame = PandasDataframeSampler.compress(
        df_measures=df_the_input,
        lst_key_columns=key_columns,
        lst_value_columns=value_columns,
        lst_ordering_columns=ordering_columns,
        b_keep_last=keep_last,
    )
    _logger.info(f"After dropping duplicates, dataframe is '{df_the_return.shape}'.")

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = df_the_return
    else:
        ctx_context.obj[alias_in] = df_the_return

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
from inspect import stack

import click
from click import BadParameter

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher import DpGrapher_DashApplication

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp-graph')
@click.pass_context
@click.argument('alias', type=str)
def dp_graph_to_dash_date_predictability(ctx_context, alias):
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot graph '{alias}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively computed a date predictability bean, and labelled it with the alias "
                       f"'{alias}'.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias], PredictabilityBean):
        raise BadParameter(
            message=f"Cannot graph '{alias}', as this alias is of type '{type(ctx_context.obj[alias])}' while "
                    f"'PredictabilityBean' is expected.",
            param_hint=f"Ensure you effectively provide an instance of 'PredictabilityBean' to graph.",
            ctx=ctx_context
        )

    obj_the_data: PredictabilityBean = ctx_context.obj[alias]

    DpGrapher_DashApplication(obj_predictability=obj_the_data).graph_predictability()

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger

from dash import callback, Output, Input, html

from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_session import Session
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_p1_welcome.tab_welcome__as_div import \
    TabWelcomeAsDiv
import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_wid as wid
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_p2_explore.tab_explore__as_div import \
    TabExploreAsDiv

_logger: Logger = getLogger(__name__)


@callback(
    Output(wid.str_the_wid__content_as_main, 'children'),
    Input(wid.str_the_wid__url_as_location, 'hash'),
    prevent_initial_call=True
)
def _callback_navigate_and_load_content(url_hash: str) -> html.Div:
    """
    Callback function to update the main content.
    :param url_hash:
    :return:
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    if url_hash == '#explore':
        div_the_return: html.Div = TabExploreAsDiv()

    elif (url_hash == '#welcome') | (url_hash == ''):
        div_the_return: html.Div = TabWelcomeAsDiv()

    else:
        raise Exception(f"Unexpected url_hash '{url_hash}'.")

    if url_hash != '':
        Session.debug(debug_trace=f"Click navigation menus to update main content: {url_hash}")

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    return div_the_return



import logging
from inspect import stack
from pathlib import Path

import click
from click import BadParameter

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_persister import DatePredictabilityPersister

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp-persist')
@click.pass_context
@click.argument('alias', type=str)
@click.argument('file', type=click.Path(exists=False))
def dp_write_bean_date_predictability(ctx_context, alias, file):
    _logger.info(f"\n"
                 f"\n**************************************************************************************************"
                 f"\n**************************************************************************************************"
                 f"\nCommand dp-persist called with parameters:"
                 f"\n\t- alias: '{alias}'"
                 f"\n\t- file: '{file}'"
                 )

    if alias not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot persist '{alias}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively computed a date predictability bean, and labelled it with the alias "
                       f"'{alias}'.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias], PredictabilityBean):
        raise BadParameter(
            message=f"Cannot persist '{alias}', as this alias is of type '{type(ctx_context.obj[alias])}' while "
                    f"'PredictabilityBean' is expected.",
            param_hint=f"Ensure you effectively provide an instance of 'PredictabilityBean' to persist.",
            ctx=ctx_context
        )

    obj_the_data: PredictabilityBean = ctx_context.obj[alias]

    DatePredictabilityPersister().persist(
        obj_bean_to_persist=obj_the_data,
        p_file_path=Path(file),
    )

    _logger.info(f"\nCommand dp-persist is returning, after having persisted in a file:"
                 f"\n\t- '{alias}' bean that contains:"                 
                 f"\n\t\t- 'by measure' dataframe:"
                 f"\n\t\t\t- shape: {obj_the_data.df_by_measure.shape}"
                 f"\n\t\t\t- columns: {', '.join(obj_the_data.df_by_measure.columns)}"
                 f"\n\t\t- 'by key' dataframe:"
                 f"\n\t\t\t- shape: {obj_the_data.df_by_key.shape}"
                 f"\n\t\t\t- columns: {', '.join(obj_the_data.df_by_key.columns)}"
                 f"\n\t\t- 'historical' dataframe:"
                 f"\n\t\t\t- shape: {obj_the_data.df_historical.shape}"
                 f"\n\t\t\t- columns: {', '.join(obj_the_data.df_historical.columns)}"
                 f"\n**************************************************************************************************"
                 f"\n**************************************************************************************************"
                 f"\n")


@click.command('dp-load')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='csv')
def dp_load_bean_date_predictability(ctx_context, file, alias):
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with an instance of 'PredictabilityBean' from file '{file}'."
        )

    obj_the_bean: PredictabilityBean = DatePredictabilityPersister().load(
            p_file_path=Path(file),
    )

    # We update the context data store
    ctx_context.obj[alias] = obj_the_bean
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "load_predictability_bean"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path

from dash import Dash
import dash.html as html
import dash_bootstrap_components as dbc
from dash import dcc

import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_wid as wid
from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_session import Session
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_debug.tab_debug__as_div import TabDebug
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_p1_welcome.tab_welcome__as_div \
    import TabWelcomeAsDiv


class DpGrapher_DashApplication:
    _logger: Logger = getLogger(__name__)

    # CONSTANTS
    # Link to the directory "assets", where web site's assets are stored (css, pictures, etc)
    str_the_relative_path_to_assets: str = '../../../../../../../assets/'
    # The application name
    str_the_application_title: str = "com.enovation"

    # The Dash application
    _dash_app: Dash = None

    # Debug
    _b_debug: bool = False

    def __init__(
            self,
            obj_predictability: PredictabilityBean,
            b_debug: bool = False,
    ):
        """
        We instantiate the dash application, along with the different widgets, including their callbacks
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # ##############################################################################################################
        # Callbacks should be imported here... Otherwise, we face "unknown id" as callbacks get registered even though
        # the widgets are not used...
        # noinspection PyUnresolvedReferences
        import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_callbacks
        # ##############################################################################################################

        self._b_debug = b_debug

        Session.set_predictability_bean(obj_predictability_bean=obj_predictability)
        Session.debug(
            f"'{stack()[0].filename} - {stack()[0].function}' loaded a predictability bean with the following "
            f"characteristics:"
            f"\n\t- df_by_key:"
            f"\n\t\t- shape: {obj_predictability.df_by_key.shape}"
            f"\n\t\t- columns: {obj_predictability.df_by_key.columns}"
            f"\n\t- df_by_measure:"
            f"\n\t\t- shape: {obj_predictability.df_by_measure.shape}"
            f"\n\t\t- columns: {obj_predictability.df_by_measure.columns}"
            f"\n\t- df_historical:"
            f"\n\t\t- shape: {obj_predictability.df_historical.shape}"
            f"\n\t\t- columns: {obj_predictability.df_historical.columns}"
        )

        # We check that the directory "assets" exists...
        if not Path(__file__, "../", self.str_the_relative_path_to_assets).resolve().is_dir():
            raise Exception(
                f'When trying to instantiate the Dash application, we could not find the directory assets":'
                f'\n\t- From current file path "{Path(__file__)}"'
                f'\n\t- Adding relative path "{self.str_the_relative_path_to_assets}"'
                f'\n\t- Leading to non existing directory '
                f'"{Path(__file__, "../", self.str_the_relative_path_to_assets).resolve()}'
            )

        # We instantiate the Dash application
        self._dash_app = Dash(
            name=__name__,
            assets_folder=os.path.join(os.path.dirname(__file__), self.str_the_relative_path_to_assets),
            external_stylesheets=[
                dbc.themes.JOURNAL,
            ],
            title=self.str_the_application_title,
            suppress_callback_exceptions=True
        )

        # We instantiate the main widgets
        self._widget_header__as_navbar = self._instantiate_widget__enov_header__as_navbar()
        self._widget_footer__as_footer = self._instantiate_widget__enov_footer__as_footer()
        self._widget_content__debug__as_div = TabDebug()

        # We eventually instantiate the overall layout to load
        self._dash_app.layout = self._instantiate_widget__enov_layout__as_div()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def _instantiate_widget__enov_layout__as_div(self) -> html.Div:
        """
        Return the overall layout for the application.

        The produced HTML is of this kind:
        TODO
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We instantiate the navbar
        div_the_return: html.Div = html.Div([

            # To get the url loaded
            dcc.Location(id=wid.str_the_wid__url_as_location),

            # The navigation bar at the top of the screen
            self._widget_header__as_navbar,

            # The main page in a container
            dbc.Container(
                children=[
                    html.Main(
                        id=wid.str_the_wid__content_as_main,
                        className="container",
                        children=TabWelcomeAsDiv()
                    ),
                    self._widget_content__debug__as_div,
                ],
            ),

            # The footer
            self._widget_footer__as_footer
        ])
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return div_the_return

    def _instantiate_widget__enov_header__as_navbar(self) -> dbc.Navbar:
        """
        Return the navigation bar, along with the callback function to load the enov-content__as_main part.

        The produced HTML is of this kind:
        <nav id="enov-header__as_navbar" class="fixed-top navbar navbar-expand-md navbar-dark bg-dark">
            <div class="nav-item">
                <a href="https://pypi.org/project/com-enovation/">
                    <div class="align-items-center row">
                        <div class="col">
                            <img height="30px" src="/assets/images/com-enovation.square.png" title="com.enovation">
                        </div>
                        <div class="col">
                            <span class="navbar-brand">eNOVAtion</span>
                        </div>
                    </div>
                </a>
            </div>
            <div class="nav-item">
                <a href="#welcome" class="nav-link">welcome</a>
            </div>
            <div class="nav-item">
                <a href="#explore" class="nav-link">explore</a>
            </div>
            <div class="nav-item">
                <a class="nav-link disabled" disabled="">The default dp_grapher for date predictability module...</a>
            </div>
        </nav>
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We instantiate the navbar
        nav_the_return: dbc.Navbar = dbc.Navbar(
            children=[
                dbc.NavItem(
                    html.A(
                        dbc.Row(
                            [
                                dbc.Col(
                                    html.Img(
                                        src=self._dash_app.get_asset_url("images/com-enovation.square.png"),
                                        height="30px",
                                        title=self.str_the_application_title
                                    )
                                ),
                                dbc.Col(dbc.NavbarBrand("eNOVAtion")),
                            ],
                            align="center",
                        ),
                        href="https://pypi.org/project/com-enovation/"
                    )
                ),
                dbc.NavItem(dbc.NavLink("welcome", href="#welcome")),
                dbc.NavItem(dbc.NavLink("explore", href="#explore")),
                dbc.NavItem(dbc.NavLink("The default dp_grapher for the date predictability module...", disabled=True))
            ],
            color="dark",
            dark=True,
            id=wid.str_the_wid__header_as_navbar,
            className="fixed-top"
        )

        return nav_the_return

    def _instantiate_widget__enov_footer__as_footer(self) -> html.Footer:
        """
        Return the footer such as:
        <footer class="footer" id="enov-footer__as_footer">
            <div class="container">
                <span class="text-muted">Copyright © 2022 Jean-Sébastien Guillard. All rights reserved.</span>
            </div>
        </footer>
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We instantiate the footer
        footer_the_return: html.Footer = html.Footer(
            html.Div(
                [
                    html.Span(
                        [
                            "Copyright © 2022 Jean-Sébastien Guillard. All rights reserved."
                        ],
                        className="text-muted"
                    )
                ],
                className="container",
            ),
            className="footer",
            id=wid.str_the_wid__footer_as_footer,
        )
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return footer_the_return

    def graph_predictability(
            self,
    ):
        """
        Main function call to instantiate and launch dash application.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
        self._dash_app.run_server(debug=True, use_reloader=False)
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



# Widgets IDs, aka wid
str_the_wid__url_as_location: str = "dp-url__as_location"
str_the_wid__session_id_as_div: str = "dp-session-id__as_div"
str_the_wid__content_as_main: str = "dp-content__as_main"
str_the_wid__header_as_navbar: str = "dp-header__as_navbar"
str_the_wid__footer_as_footer: str = "dp-footer__as_footer"

# Widgets for debug
str_the_wid__debug_as_div: str = "dp-content__debug__as_div"
str_the_wid__debug_as_samp: str = "dp-content__debug__as_samp"
str_the_wid__debug_as_interval: str = "dp-content__debug__as_interval"

# #####################################################################################################################
# #####################################################################################################################
# Widgets for exploring
# #####################################################################################################################
# #####################################################################################################################

# the drop down to select a key
str_the_wid__explore__key__as_drop_down: str = "dp-content__explore__key__as_drop_down"

# the div containing the dash widget to graph the selected key
str_the_wid__explore__graph__as_div: str = "dp-content__explore__graph__as_div"

# the dash graph widget
str_the_wid__explore__graph__as_graph: str = "dp-content__explore__graph__as_graph"

# the div containing the table displaying selected key's details
str_the_wid__explore__table__as_div: str = "dp-content__explore__table__as_div"

str_the_wid__explore__table__prefix_to_access_columns: str = "dp-content__explore__table__"
str_the_wid__explore__table__suffix_to_access_columns: str = "__as_td"

str_the_wid__explore__table__key__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "key__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__measure_count__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "measure_count__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__date_first__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "date_first__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__date_late__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "date_last__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__measure_min__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "measure_min__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__measure_max__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "measure_max__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__measure_first__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "measure_first__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__measure_last__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "measure_last__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns
str_the_wid__explore__table__predictability_last__as_td: str =\
    str_the_wid__explore__table__prefix_to_access_columns + "predictability_last__as_td" +\
    str_the_wid__explore__table__suffix_to_access_columns



from inspect import stack
from logging import Logger, getLogger
from typing import Union


class Session:
    """
    The class session is used to persist on the server side data related to the user's session.

    Among other things, the class is used to keep track for:
    - The debug trace: whenever we want to debug an information, we can call the function "debug", which will append
      the message and display the whole trace into the web browser (if debug option is activated)
    """

    _logger: Logger = getLogger(__name__)

    # ################################################################################################################ #
    # The logged message, to be printed in the debug                                                                   #
    # ################################################################################################################ #

    @property
    def lst_str_debug_trace(self) -> list[str]:
        return self._lst_str_debug_trace

    @property
    def dash_application(self) -> object:
        return self._dash_application

    @property
    def dict_sheets(self) -> dict:
        return self._dash_application.dict_sheets

    @property
    def dict_context(self) -> dict:
        return self._dash_application.dict_context

    @property
    def dict_config(self) -> dict:
        return self._dash_application.dict_config

    def __init__(
            self,
            dash_application
    ):
        # The dictionary is organized by session id, and then provide the list of all debut messages
        self._lst_str_debug_trace: list[str] = []
        self._dash_application = dash_application

    def debug(self, debug_trace: Union[str, list[str]]):
        """
        Function that logs new messages to be printed.

        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
                           f"\n\tdebug_trace type : '{type(debug_trace)}'"
                           f"\n\tdebug_trace value: '{debug_trace}'.")

        lst_the_list_of_traces: list[str]

        if isinstance(debug_trace, str):
            self.lst_str_debug_trace.append(debug_trace)
        else:
            self._lst_str_debug_trace = self.lst_str_debug_trace

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
